"""
Post-processing algorithms modify predictions to be more fair (predictions in,
predictions out).
"""
from logging import warning

import pandas as pd
from sklearn.base import BaseEstimator, MetaEstimatorMixin, clone
from sklearn.model_selection import train_test_split
from sklearn.utils.metaestimators import available_if

from aif360.sklearn.postprocessing.calibrated_equalized_odds import CalibratedEqualizedOdds
from aif360.sklearn.postprocessing.reject_option_classification import RejectOptionClassifier, RejectOptionClassifierCV


class PostProcessingMeta(BaseEstimator, MetaEstimatorMixin):
    """A meta-estimator which wraps a given estimator with a post-processing
    step.

    The post-processor trains on a separate training set from the estimator to
    prevent leakage.

    Note:
        Because of the dataset splitting, if a Pipeline is necessary it should
        be used as the input to this meta-estimator not the other way around.

    Attributes:
        estimator_: Fitted estimator.
        postprocessor_: Fitted postprocessor.
        classes_ (array, shape (n_classes,)): Class labels from `estimator_`.
    """

    def __init__(self, estimator, postprocessor, *, prefit=False, val_size=0.25,
                 **options):
        """
        Args:
            estimator (sklearn.BaseEstimator): Original estimator.
            postprocessor: Post-processing algorithm.
            prefit (bool): If ``True``, it is assumed that estimator has been
                fitted already and all data is used to train postprocessor.
            val_size (int or float): Size of validation set used to fit the
                postprocessor. The estimator fits on the remainder of the
                training set.
                See :func:`~sklearn.model_selection.train_test_split` for
                details.
            **options: Keyword options passed through to
                :func:`~sklearn.model_selection.train_test_split`.
                Note: 'train_size' and 'test_size' will be ignored in favor of
                'val_size'.
        """
        self.estimator = estimator
        self.postprocessor = postprocessor
        self.prefit = prefit
        self.val_size = val_size
        self.options = options

    @property
    def _estimator_type(self):
        return self.postprocessor._estimator_type

    @property
    def classes_(self):
        """Class labels from the base estimator."""
        return self.estimator_.classes_

    def fit(self, X, y, sample_weight=None, **fit_params):
        """Splits the training samples with
        :func:`~sklearn.model_selection.train_test_split` and uses the resultant
        'train' portion to train the estimator. Then the estimator predicts on
        the 'test' portion of the split data and the post-processor is trained
        with those prediction-ground-truth target pairs.

        Args:
            X (array-like): Training samples.
            y (pandas.Series): Training labels.
            sample_weight (array-like, optional): Sample weights.
            **fit_params: Parameters passed to the post-processor ``fit()``
                method. Note: these do not need to be prefixed with ``__``
                notation.

        Returns:
            self
        """
        self.postprocessor_ = clone(self.postprocessor)
        self.estimator_ = self.estimator if self.prefit else clone(self.estimator)

        try:
            use_proba = self.postprocessor._get_tags()['requires_proba']
        except KeyError:
            raise TypeError("`postprocessor` (type: {}) does not have a "
                            "'requires_proba' tag.".format(type(self.estimator)))
        if use_proba and not hasattr(self.estimator, 'predict_proba'):
            raise TypeError("`estimator` (type: {}) does not implement method "
                            "`predict_proba()`.".format(type(self.estimator)))

        if self.prefit:
            if len(self.options):
                warning("Splitting options were passed but prefit is True so "
                        "these are ignored.")
            y_score = (self.estimator_.predict_proba(X) if use_proba else
                       self.estimator_.predict(X))
            y_score = pd.DataFrame(y_score, index=X.index).squeeze('columns')
            fit_params = fit_params.copy()
            fit_params.update(labels=self.estimator_.classes_)
            self.postprocessor_.fit(y_score, y, sample_weight=sample_weight,
                                    **fit_params)
            return self

        if 'train_size' in self.options or 'test_size' in self.options:
            warning("'train_size' and 'test_size' are ignored in favor of "
                    "'val_size'")
        options_ = self.options.copy()
        options_['test_size'] = self.val_size
        if 'train_size' in options_:
            del options_['train_size']

        if sample_weight is not None:
            X_est, X_post, y_est, y_post, sw_est, sw_post = train_test_split(
                    X, y, sample_weight, **options_)
            self.estimator_.fit(X_est, y_est, sample_weight=sw_est)
        else:
            X_est, X_post, y_est, y_post = train_test_split(X, y, **options_)
            self.estimator_.fit(X_est, y_est)

        y_score = (self.estimator_.predict_proba(X_post) if use_proba else
                   self.estimator_.predict(X_post))
        y_score = pd.DataFrame(y_score, index=X_post.index).squeeze('columns')
        fit_params = fit_params.copy()
        fit_params.update(labels=self.estimator_.classes_)
        self.postprocessor_.fit(y_score, y_post, sample_weight=sw_post
                                if sample_weight is not None else None,
                                **fit_params)
        return self

    @available_if(lambda self: hasattr(self.postprocessor_, "predict"))
    def predict(self, X):
        """Predict class labels for the given samples.

        First, runs ``self.estimator_.predict()`` (or ``predict_proba()`` if
        required) then returns the post-processed output from those predictions.

        Args:
            X (pandas.DataFrame): Test samples.

        Returns:
            numpy.ndarray: Predicted class label per sample.
        """
        use_proba = self.postprocessor_._get_tags()['requires_proba']
        y_score = (self.estimator_.predict_proba(X) if use_proba else
                   self.estimator_.predict(X))
        y_score = pd.DataFrame(y_score, index=X.index).squeeze('columns')
        return self.postprocessor_.predict(y_score)

    @available_if(lambda self: hasattr(self.postprocessor_, "predict_proba"))
    def predict_proba(self, X):
        """Probability estimates.

        First, runs ``self.estimator_.predict()`` (or ``predict_proba()`` if
        required) then returns the post-processed output from those predictions.

        The returned estimates for all classes are ordered by the label of
        classes.

        Args:
            X (pandas.DataFrame): Test samples.

        Returns:
            numpy.ndarray: Returns the probability of the sample for each class
            in the model, where classes are ordered as they are in
            ``self.classes_``.
        """
        use_proba = self.postprocessor_._get_tags()['requires_proba']
        y_score = (self.estimator_.predict_proba(X) if use_proba else
                   self.estimator_.predict(X))
        y_score = pd.DataFrame(y_score, index=X.index).squeeze('columns')
        return self.postprocessor_.predict_proba(y_score)

    @available_if(lambda self: hasattr(self.postprocessor_, "predict_log_proba"))
    def predict_log_proba(self, X):
        """Log of probability estimates.

        First, runs ``self.estimator_.predict()`` (or ``predict_proba()`` if
        required) then returns the post-processed output from those predictions.

        The returned estimates for all classes are ordered by the label of
        classes.

        Args:
            X (pandas.DataFrame): Test samples.

        Returns:
            array: Returns the log-probability of the sample for each class in
            the model, where classes are ordered as they are in
            ``self.classes_``.
        """
        use_proba = self.postprocessor_._get_tags()['requires_proba']
        y_score = (self.estimator_.predict_proba(X) if use_proba else
                   self.estimator_.predict(X))
        y_score = pd.DataFrame(y_score, index=X.index).squeeze('columns')
        return self.postprocessor_.predict_log_proba(y_score)

    @available_if(lambda self: hasattr(self.postprocessor_, "score"))
    def score(self, X, y, sample_weight=None):
        """Returns the output of the post-processor's score function on the
        given test data and labels.

        First, runs ``self.estimator_.predict()`` (or ``predict_proba()`` if
        required) then gets the post-processed output from those predictions and
        scores it.

        Args:
            X (pandas.DataFrame): Test samples.
            y (array-like): True labels for X.
            sample_weight (array-like, optional): Sample weights.

        Returns:
            float: Score value.
        """
        use_proba = self.postprocessor_._get_tags()['requires_proba']
        y_score = (self.estimator_.predict_proba(X) if use_proba else
                   self.estimator_.predict(X))
        y_score = pd.DataFrame(y_score, index=X.index).squeeze('columns')
        if sample_weight is None:
            return self.postprocessor_.score(y_score, y)
        return self.postprocessor_.score(y_score, y,
                                         sample_weight=sample_weight)


__all__ = [
    'CalibratedEqualizedOdds', 'PostProcessingMeta', 'RejectOptionClassifier',
    'RejectOptionClassifierCV'
]



from itertools import product

import numpy as np
import pandas as pd
import pytest
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split
from sklearn.metrics import make_scorer

from aif360.algorithms.postprocessing import RejectOptionClassification
from aif360.sklearn.postprocessing import RejectOptionClassifier, RejectOptionClassifierCV, PostProcessingMeta
from aif360.sklearn.metrics import generalized_entropy_error


@pytest.fixture(scope='module')
def log_reg_probs(new_adult):
    """Train a LogisticRegression model and return predicted probabilities."""
    X, y, sample_weight = new_adult
    lr = LogisticRegression(solver='lbfgs', max_iter=500)
    lr.fit(X, y, sample_weight=sample_weight)
    return lr.predict_proba(X)

@pytest.fixture(scope='module')
def old_ROC(old_adult, log_reg_probs):
    """Fit old ROC."""
    old_adult_pred = old_adult.copy()
    old_adult_pred.scores = log_reg_probs[:, [1]]

    ROC = RejectOptionClassification(unprivileged_groups=[{'sex': 0}],
            privileged_groups=[{'sex': 1}], low_class_thresh=0.1,
            high_class_thresh=0.9, num_class_thresh=9, num_ROC_margin=9,
            metric_name='Statistical parity difference', metric_ub=0.1,
            metric_lb=-0.1)
    return ROC.fit(old_adult, old_adult_pred)

@pytest.fixture(scope='module')
def new_ROC(new_adult, log_reg_probs):
    """Fit new ROC."""
    _, y, _ = new_adult
    y_pred = pd.DataFrame(log_reg_probs, index=y.index)

    ROC = RejectOptionClassifierCV('sex', scoring='statistical_parity',
                                   step=0.1, n_jobs=-1)
    return ROC.fit(y_pred, y)

def test_rej_opt_clf_fit(new_ROC, old_ROC):
    """Test RejectOptionClassifierCV fit matches old."""
    assert np.isclose(new_ROC.best_estimator_.threshold,
                      old_ROC.classification_threshold, atol=0.05)
    assert np.isclose(new_ROC.best_estimator_.margin, old_ROC.ROC_margin, atol=0.05)

def test_rej_opt_clf_predict(new_adult, old_adult, log_reg_probs, old_ROC):
    """Test RejectOptionClassifier predict matches old."""
    _, y, _ = new_adult
    old_adult_pred = old_adult.copy()
    old_adult_pred.scores = log_reg_probs[:, [1]]

    threshold = old_ROC.classification_threshold
    margin = old_ROC.ROC_margin
    ROC = RejectOptionClassifier('sex', threshold=threshold, margin=margin)
    y_pred = pd.DataFrame(log_reg_probs, index=y.index)
    y_postpred = ROC.fit_predict(y_pred, y)

    assert np.allclose(y_postpred, old_ROC.predict(old_adult_pred).labels.ravel())

def test_rej_opt_clf_predict_proba(new_adult, log_reg_probs, new_ROC):
    """Test RejectOptionClassifier predict_proba + threshold matches predict."""
    _, y, _ = new_adult
    y_pred = pd.DataFrame(log_reg_probs, index=y.index)
    y_postpred = new_ROC.predict(y_pred)
    y_postproba = new_ROC.predict_proba(y_pred)[:, 1]

    threshold = new_ROC.best_estimator_.threshold
    assert all((y_postproba > threshold).astype(int) == y_postpred)

    # test pos_label=0
    y_pred = pd.DataFrame(log_reg_probs[:, ::-1], index=y.index)
    new_ROC.fit(y_pred, 1-y, pos_label=0)

    assert threshold == new_ROC.best_estimator_.threshold
    assert all(y_postpred == 1-new_ROC.predict(y_pred))
    assert all(y_postproba == new_ROC.predict_proba(y_pred)[:, 0])

@pytest.mark.filterwarnings('error', category=RuntimeWarning)
def test_rej_opt_clf_errors():
    y_proba = np.array([[0.5, 0.5], [0.3, 0.7], [0.55, 0.45]])
    # y_true = pd.Series([0, 0, 0], index=[1, 1, 0]).rename_axis(index='prot')
    y_true = pd.DataFrame([[0, 1], [0, 1], [0, 0]], columns=['feat', 'prot'])
    y_true = y_true.set_index('prot', append=True).squeeze()
    roc = RejectOptionClassifier('prot')
    with pytest.raises(ValueError):  # binary classification
        roc.fit(y_proba, y_true)
    with pytest.raises(TypeError):  # missing prot attr
        roc.fit(y_proba, y_true, labels=[0, 1]).predict(y_proba)
    y_proba = pd.DataFrame(y_proba, index=y_true.index)
    y_postpred = roc.fit_predict(y_proba, labels=[0, 1])
    assert (y_postpred == [0, 1, 1]).all()

    cv = RejectOptionClassifierCV('prot', scoring='statistical_parity',
                                  step=0.1, cv=3)
    with pytest.raises(RuntimeWarning):  # sample_weight ignored warning
        cv.fit(y_proba, y_true, sample_weight=[10, 1, 1])

def test_rej_opt_clf_postproc_meta(new_adult, new_ROC):
    """Test PostProcessingMeta pipeline with RejectOptionClassifierCV."""
    X, y, sample_weight = new_adult
    lr = LogisticRegression(solver='lbfgs', max_iter=500)
    lr.fit(X, y, sample_weight=sample_weight)
    pp = PostProcessingMeta(lr,
            RejectOptionClassifierCV('sex', scoring='statistical_parity',
                                     step=0.1, n_jobs=-1),
            prefit=True, random_state=1234)
    pp.fit(X, y)
    assert pp.postprocessor_.best_params_ == new_ROC.best_params_
    assert (pp.postprocessor_.cv_results_['mean_test_bal_acc']
         == new_ROC.cv_results_['mean_test_bal_acc']).all()

    pp.set_params(prefit=False).fit(X, y, sample_weight=sample_weight)

    X_train, X_test, y_train, y_test, sw_train, sw_test = train_test_split(X, y,
            sample_weight, train_size=0.75, random_state=1234)
    lr.fit(X_train, y_train, sample_weight=sw_train)
    y_pred = pd.DataFrame(lr.predict_proba(X_test), index=y_test.index)
    new_ROC.fit(y_pred, y_test)
    assert (lr.coef_ == pp.estimator_.coef_).all()
    assert pp.postprocessor_.best_params_ == new_ROC.best_params_
    assert (pp.postprocessor_.cv_results_['mean_test_bal_acc']
         == new_ROC.cv_results_['mean_test_bal_acc']).all()

@pytest.mark.parametrize(
    "prot_attr, scoring",
    product(['race', 'sex'], ['statistical_parity', 'average_odds',
                              'equal_opportunity', 'disparate_impact'])
)
def test_rej_opt_clf_scoring(prot_attr, scoring, new_adult):
    """Test all scoring options work."""
    X, y, sample_weight = new_adult
    pp = PostProcessingMeta(LogisticRegression(solver='lbfgs', max_iter=500),
            RejectOptionClassifierCV(prot_attr, scoring=scoring, step=0.1,
                                     refit=scoring, n_jobs=-1))
    pp.fit(X, y, sample_weight=sample_weight)

    if scoring == 'disparate_impact':
        assert pp.score(X, y) >= 0.8
    else:
        assert pp.score(X, y) >= -0.1

def test_rej_opt_clf_custom_scoring(new_adult):
    X, y, sample_weight = new_adult
    scoring = make_scorer(generalized_entropy_error, greater_is_better=False)
    pp = PostProcessingMeta(LogisticRegression(solver='lbfgs', max_iter=500),
            RejectOptionClassifierCV('sex', scoring=scoring, step=0.1,
                                     refit=False, n_jobs=-1))
    pp.fit(X, y, sample_weight=sample_weight)
    res = pd.DataFrame(pp.postprocessor_.cv_results_)
    assert not res.isna().any(axis=None)



"""
The code for ExponentiatedGradientReduction wraps the source class
fairlearn.reductions.ExponentiatedGradient
available in the https://github.com/fairlearn/fairlearn library
licensed under the MIT Licencse, Copyright Microsoft Corporation
"""
try:
    import fairlearn.reductions as red
except ImportError as error:
    from logging import warning
    warning("{}: ExponentiatedGradientReduction will be unavailable. To install, run:\n"
            "pip install 'aif360[Reductions]'".format(error))
from sklearn.base import BaseEstimator, ClassifierMixin, clone
from sklearn.preprocessing import LabelEncoder


class ExponentiatedGradientReduction(BaseEstimator, ClassifierMixin):
    """Exponentiated gradient reduction for fair classification.

    Exponentiated gradient reduction is an in-processing technique that reduces
    fair classification to a sequence of cost-sensitive classification problems,
    returning a randomized classifier with the lowest empirical error subject to
    fair classification constraints [#agarwal18]_.

    References:
        .. [#agarwal18] `A. Agarwal, A. Beygelzimer, M. Dudik, J. Langford, and
           H. Wallach, "A Reductions Approach to Fair Classification,"
           International Conference on Machine Learning, 2018.
           <https://arxiv.org/abs/1803.02453>`_
    """
    def __init__(self,
                 prot_attr,
                 estimator,
                 constraints,
                 eps=0.01,
                 max_iter=50,
                 nu=None,
                 eta0=2.0,
                 run_linprog_step=True,
                 drop_prot_attr=True):
        """
        Args:
            prot_attr: String or array-like column indices or column names of
                protected attributes.
            estimator: An estimator implementing methods
                ``fit(X, y, sample_weight)`` and ``predict(X)``, where ``X`` is
                the matrix of features, ``y`` is the vector of labels, and
                ``sample_weight`` is a vector of weights; labels ``y`` and
                predictions returned by ``predict(X)`` are either 0 or 1 -- e.g.
                scikit-learn classifiers.
            constraints (str or fairlearn.reductions.Moment): If string, keyword
                denoting the :class:`fairlearn.reductions.Moment` object
                defining the disparity constraints -- e.g., "DemographicParity"
                or "EqualizedOdds". For a full list of possible options see
                `self.model.moments`. Otherwise, provide the desired
                :class:`~fairlearn.reductions.Moment` object defining the
                disparity constraints.
            eps: Allowed fairness constraint violation; the solution is
                guaranteed to have the error within ``2*best_gap`` of the best
                error under constraint eps; the constraint violation is at most
                ``2*(eps+best_gap)``.
            max_iter: Maximum number of iterations.
            nu: Convergence threshold for the duality gap, corresponding to a
                conservative automatic setting based on the statistical
                uncertainty in measuring classification error.
            eta0: Initial setting of the learning rate.
            run_linprog_step: If True each step of exponentiated gradient is
                followed by the saddle point optimization over the convex hull
                of classifiers returned so far.
            drop_prot_attr: Boolean flag indicating whether to drop protected
                attributes from training data.
        """
        self.prot_attr = prot_attr
        self.estimator = estimator
        self.constraints = constraints
        self.eps = eps
        self.max_iter = max_iter
        self.nu = nu
        self.eta0 = eta0
        self.run_linprog_step = run_linprog_step
        self.drop_prot_attr = drop_prot_attr

    def fit(self, X, y):
        """Learns randomized model with less bias

        Args:
            X (pandas.DataFrame): Training samples.
            y (array-like): Training labels.

        Returns:
            self
        """
        self.estimator_ = clone(self.estimator)

        moments = {
            "DemographicParity": red.DemographicParity,
            "EqualizedOdds": red.EqualizedOdds,
            "TruePositiveRateParity": red.TruePositiveRateParity,
            "FalsePositiveRateParity": red.FalsePositiveRateParity,
            "ErrorRateParity": red.ErrorRateParity,
        }
        if isinstance(self.constraints, str):
            if self.constraints not in moments:
                raise ValueError(f"Constraint not recognized: {self.constraints}")
            self.moment_ = moments[self.constraints]()
        elif isinstance(self.constraints, red.Moment):
            self.moment_ = self.constraints
        else:
            raise ValueError("constraints must be a string or Moment object.")

        self.model_ = red.ExponentiatedGradient(self.estimator_, self.moment_,
                eps=self.eps, max_iter=self.max_iter, nu=self.nu, eta0=self.eta0)

        A = X[self.prot_attr]

        if self.drop_prot_attr:
            X = X.drop(self.prot_attr, axis=1)

        le = LabelEncoder()
        y = le.fit_transform(y)
        self.classes_ = le.classes_

        self.model_.fit(X, y, sensitive_features=A)

        return self


    def predict(self, X):
        """Predict class labels for the given samples.
        Args:
            X (pandas.DataFrame): Test samples.
        Returns:
            numpy.ndarray: Predicted class label per sample.
        """
        if self.drop_prot_attr:
            X = X.drop(self.prot_attr, axis=1)

        return self.classes_[self.model_.predict(X)]


    def predict_proba(self, X):
        """Probability estimates.

        The returned estimates for all classes are ordered by the label of
        classes.

        Args:
            X (pandas.DataFrame): Test samples.

        Returns:
            numpy.ndarray: Returns the probability of the sample for each class
            in the model, where classes are ordered as they are in
            ``self.classes_``.
        """
        if self.drop_prot_attr:
            X = X.drop(self.prot_attr, axis=1)

        return self.model_._pmf_predict(X)



import numpy as np
from aif360.detectors.mdss.ScoringFunctions.ScoringFunction import ScoringFunction


def bisection_q_mle(score_function: ScoringFunction, observed_sum: float, probs: np.array, **kwargs):
    """
    Computes the q which maximizes score (q_mle).
    Computes q for which slope dscore/dq = 0, using the fact that slope is monotonically decreasing.
    q_mle is computed via bisection.
    This works if the score, as a function of q, is concave.
    So the slope is monotonically decreasing, and q_mle is the unique value for which slope = 0.

    :param observed_sum: sum of observed binary outcomes for all i
    :param probs: predicted probabilities p_i for each data element i
    :return: q MLE
    """
    q_temp_min = 1e-3
    q_temp_max = 1e3

    while np.abs(q_temp_max - q_temp_min) > 1e-3:
        q_temp_mid = (q_temp_min + q_temp_max) / 2

        if np.sign(score_function.q_dscore(observed_sum, probs, q_temp_mid)) > 0:
            q_temp_min = q_temp_min + (q_temp_max - q_temp_min) / 2
        else:
            q_temp_max = q_temp_max - (q_temp_max - q_temp_min) / 2

    q = (q_temp_min + q_temp_max) / 2
    
    direction = None
    if 'direction' in kwargs:
        direction = kwargs['direction']
        
    if ((direction == 'positive') & (q < 1)) | ((direction == 'negative') & (q > 1)):
        return 1
        
    return q

def bisection_q_min(score_function: ScoringFunction, observed_sum: float, probs: np.array, penalty: float, q_mle: float, temp_min = 1e-3, **kwargs):
    """
    Compute q for which score = 0, using the fact that score is monotonically increasing for q > q_mle.
    q_max is computed via binary search.
    This works because the score, as a function of q, is concave.

    :param observed_sum: sum of observed binary outcomes for all i
    :param probs: predicted probabilities p_i for each data element i
    :param penalty: penalty term. should be positive
    :param q_mle: q maximum likelihood
    :return: the root on the LHS of qmle
    """
    q_temp_min = temp_min
    q_temp_max = q_mle

    while np.abs(q_temp_max - q_temp_min) > 1e-3:
        q_temp_mid = (q_temp_min + q_temp_max) / 2

        if np.sign(score_function.score(observed_sum, probs, penalty, q_temp_mid)) > 0:
            q_temp_max = q_temp_max - (q_temp_max - q_temp_min) / 2
        else:
            q_temp_min = q_temp_min + (q_temp_max - q_temp_min) / 2

    return (q_temp_min + q_temp_max) / 2

def bisection_q_max(score_function: ScoringFunction, observed_sum: float, probs: np.array, penalty: float, q_mle: float, temp_max = 1e3, **kwargs):
    """
    Compute q for which score = 0, using the fact that score is monotonically decreasing for q > q_mle.
    q_max is computed via binary search.
    This works because the score, as a function of q, is concave.

    :param observed_sum: sum of observed binary outcomes for all i
    :param probs: predicted probabilities p_i for each data element i
    :param penalty: penalty term. should be positive
    :param q_mle: q maximum likelihood
    :return: the root on the RHS of qmle
    """
    q_temp_min = q_mle
    q_temp_max = temp_max

    while np.abs(q_temp_max - q_temp_min) > 1e-3:
        q_temp_mid = (q_temp_min + q_temp_max) / 2

        if np.sign(score_function.score(observed_sum, probs, penalty, q_temp_mid)) > 0:
            q_temp_min = q_temp_min + (q_temp_max - q_temp_min) / 2
        else:
            q_temp_max = q_temp_max - (q_temp_max - q_temp_min) / 2

    return (q_temp_min + q_temp_max) / 2

def direction_assertions(direction: str, q_min: float, q_max: float):
    """
    Does some sanity checks to see if the q_mle exists for the given direction.
    """
    exist = 1
    if direction == 'positive':
        if q_max < 1:
            exist = 0
        elif q_min < 1:
            q_min = 1
    else:
        if q_min > 1:
            exist = 0
        elif q_max > 1:
            q_max = 1

    return exist, q_min, q_max


import numpy as np

from aif360.datasets import StructuredDataset


class BinaryLabelDataset(StructuredDataset):
    """Base class for all structured datasets with binary labels."""

    def __init__(self, favorable_label=1., unfavorable_label=0., **kwargs):
        """
        Args:
            favorable_label (float): Label value which is considered favorable
                (i.e. "positive").
            unfavorable_label (float): Label value which is considered
                unfavorable (i.e. "negative").
            **kwargs: StructuredDataset arguments.
        """
        self.favorable_label = float(favorable_label)
        self.unfavorable_label = float(unfavorable_label)

        super(BinaryLabelDataset, self).__init__(**kwargs)

    def validate_dataset(self):
        """Error checking and type validation.

        Raises:
            ValueError: `labels` must be shape [n, 1].
            ValueError: `favorable_label` and `unfavorable_label` must be the
                only values present in `labels`.
        """
        # fix scores before validating
        if np.all(self.scores == self.labels):
            self.scores = (self.scores == self.favorable_label).astype(np.float64)

        super(BinaryLabelDataset, self).validate_dataset()

        # =========================== SHAPE CHECKING ===========================
        # Verify if the labels are only 1 column
        if self.labels.shape[1] != 1:
            raise ValueError("BinaryLabelDataset only supports single-column "
                "labels:\n\tlabels.shape = {}".format(self.labels.shape))

        # =========================== VALUE CHECKING ===========================
        # Check if the favorable and unfavorable labels match those in the dataset
        if (not set(self.labels.ravel()) <=
                set([self.favorable_label, self.unfavorable_label])):
            raise ValueError("The favorable and unfavorable labels provided do "
                             "not match the labels in the dataset.")



# Contains the helper functions for the optim_preproc class
import numpy as np
import pandas as pd


def get_distortion_adult(vold, vnew):
    """Distortion function for the adult dataset. We set the distortion
    metric here. See section 4.3 in supplementary material of
    http://papers.nips.cc/paper/6988-optimized-pre-processing-for-discrimination-prevention
    for an example

    Note:
        Users can use this as templates to create other distortion functions.

    Args:
        vold (dict) : {attr:value} with old values
        vnew (dict) : dictionary of the form {attr:value} with new values

    Returns:
        d (value) : distortion value
    """

    # Define local functions to adjust education and age
    def adjustEdu(v):
        if v == '>12':
            return 13
        elif v == '<6':
            return 5
        else:
            return int(v)

    def adjustAge(a):
        if a == '>=70':
            return 70.0
        else:
            return float(a)

    def adjustInc(a):
        if a == "<=50K":
            return 0
        elif a == ">50K":
            return 1
        else:
            return int(a)

    # value that will be returned for events that should not occur
    bad_val = 3.0

    # Adjust education years
    eOld = adjustEdu(vold['Education Years'])
    eNew = adjustEdu(vnew['Education Years'])

    # Education cannot be lowered or increased in more than 1 year
    if (eNew < eOld) | (eNew > eOld+1):
        return bad_val

    # adjust age
    aOld = adjustAge(vold['Age (decade)'])
    aNew = adjustAge(vnew['Age (decade)'])

    # Age cannot be increased or decreased in more than a decade
    if np.abs(aOld-aNew) > 10.0:
        return bad_val

    # Penalty of 2 if age is decreased or increased
    if np.abs(aOld-aNew) > 0:
        return 2.0

    # Adjust income
    incOld = adjustInc(vold['Income Binary'])
    incNew = adjustInc(vnew['Income Binary'])

    # final penalty according to income
    if incOld > incNew:
        return 1.0
    else:
        return 0.0


def get_distortion_german(vold, vnew):
    """Distortion function for the german dataset. We set the distortion
    metric here. See section 4.3 in supplementary material of
    http://papers.nips.cc/paper/6988-optimized-pre-processing-for-discrimination-prevention
    for an example

    Note:
        Users can use this as templates to create other distortion functions.

    Args:
        vold (dict) : {attr:value} with old values
        vnew (dict) : dictionary of the form {attr:value} with new values

    Returns:
        d (value) : distortion value
    """

    # Distortion cost
    distort = {}
    distort['credit_history'] = pd.DataFrame(
                                {'None/Paid': [0., 1., 2.],
                                'Delay':      [1., 0., 1.],
                                'Other':      [2., 1., 0.]},
                                index=['None/Paid', 'Delay', 'Other'])
    distort['employment'] = pd.DataFrame(
                            {'Unemployed':    [0., 1., 2.],
                            '1-4 years':      [1., 0., 1.],
                            '4+ years':       [2., 1., 0.]},
                            index=['Unemployed', '1-4 years', '4+ years'])
    distort['savings'] = pd.DataFrame(
                            {'Unknown/None':  [0., 1., 2.],
                            '<500':           [1., 0., 1.],
                            '500+':           [2., 1., 0.]},
                            index=['Unknown/None', '<500', '500+'])
    distort['status'] = pd.DataFrame(
                            {'None':          [0., 1., 2.],
                            '<200':           [1., 0., 1.],
                            '200+':           [2., 1., 0.]},
                            index=['None', '<200', '200+'])
    distort['credit'] = pd.DataFrame(
                        {'Bad Credit' :    [0., 1.],
                         'Good Credit':    [2., 0.]},
                         index=['Bad Credit', 'Good Credit'])
    distort['sex'] = pd.DataFrame(
                        {0.0:    [0., 2.],
                         1.0:    [2., 0.]},
                         index=[0.0, 1.0])
    distort['age'] = pd.DataFrame(
                        {0.0:    [0., 2.],
                         1.0:    [2., 0.]},
                         index=[0.0, 1.0])

    total_cost = 0.0
    for k in vold:
        if k in vnew:
            total_cost += distort[k].loc[vnew[k], vold[k]]

    return total_cost

def get_distortion_compas(vold, vnew):
    """Distortion function for the compas dataset. We set the distortion
    metric here. See section 4.3 in supplementary material of
    http://papers.nips.cc/paper/6988-optimized-pre-processing-for-discrimination-prevention
    for an example

    Note:
        Users can use this as templates to create other distortion functions.

    Args:
        vold (dict) : {attr:value} with old values
        vnew (dict) : dictionary of the form {attr:value} with new values

    Returns:
        d (value) : distortion value
    """
    # Distortion cost
    distort = {}
    distort['two_year_recid'] = pd.DataFrame(
                                {'No recid.':     [0., 2.],
                                'Did recid.':     [2., 0.]},
                                index=['No recid.', 'Did recid.'])
    distort['age_cat'] = pd.DataFrame(
                            {'Less than 25':    [0., 1., 2.],
                            '25 to 45':         [1., 0., 1.],
                            'Greater than 45':  [2., 1., 0.]},
                            index=['Less than 25', '25 to 45', 'Greater than 45'])
    distort['c_charge_degree'] = pd.DataFrame(
                            {'M':   [0., 2.],
                            'F':    [1., 0.]},
                            index=['M', 'F'])
    distort['priors_count'] = pd.DataFrame(
                            {'0':           [0., 1., 2.],
                            '1 to 3':       [1., 0., 1.],
                            'More than 3':  [2., 1., 0.]},
                            index=['0', '1 to 3', 'More than 3'])
    distort['sex'] = pd.DataFrame(
                        {0.0:    [0., 2.],
                         1.0:    [2., 0.]},
                         index=[0.0, 1.0])
    distort['race'] = pd.DataFrame(
                        {0.0:    [0., 2.],
                         1.0:    [2., 0.]},
                         index=[0.0, 1.0])

    total_cost = 0.0
    for k in vold:
        if k in vnew:
            total_cost += distort[k].loc[vnew[k], vold[k]]

    return total_cost



import os.path as osp

import pandas as pd
import numpy as np
try:
    from rpy2 import robjects
    from rpy2.robjects.vectors import StrVector
    from rpy2.robjects.packages import importr
    from rpy2.robjects import pandas2ri
    from rpy2.robjects.conversion import localconverter
except ImportError as error:
    from logging import warning
    warning("{}: FairAdapt will be unavailable. To install, run:\n"
            "pip install 'aif360[FairAdapt]'".format(error))
from sklearn.base import BaseEstimator

from aif360.sklearn.utils import check_inputs, check_groups


class FairAdapt(BaseEstimator):
    """Fair Data Adaptation.

    Fairadapt is a pre-processing technique that can be used for both fair
    classification and fair regression [#plecko20]_. The method is a causal
    inference approach to bias removal and it relies on the causal graph for
    the dataset. The original implementation is in R [#plecko21]_.

    References:
        .. [#plecko20] `D. Plečko and N. Meinshausen,
           "Fair Data Adaptation with Quantile Preservation,"
           Journal of Machine Learning Research, 2020.
           <https://www.jmlr.org/papers/volume21/19-966/19-966.pdf>`_
        .. [#plecko21] `D. Plečko and N. Bennett and N. Meinshausen,
           "FairAdapt: Causal Reasoning for Fair Data Pre-processing,"
           arXiv, 2021. <https://arxiv.org/abs/2110.10200>`_

    Attributes:
        prot_attr_ (str or list(str)): Protected attribute(s) used for fair data
            adaptation.
        groups_ (array, shape (n_groups,)): A list of group labels known to the
            transformer.
    """

    def __init__(self, prot_attr, adj_mat):
        """
        Args:
            prot_attr (single label): Name of the protected attribute. Must be
                binary.
            adj_mat (array-like): A 2-dimensional array representing the
                adjacency matrix of the causal diagram of the data generating
                process. Row/column order must match `X_train`.
        """
        self.prot_attr = prot_attr
        self.adj_mat = adj_mat

        # R packages need to run FairAdapt
        pkgs = ('ranger', 'fairadapt')
        # selectively install the missing packages
        pkgs = [p for p in pkgs if not robjects.packages.isinstalled(p)]
        if len(pkgs) > 0:
            utls = robjects.packages.importr('utils')
            utls.chooseCRANmirror(ind=1)
            utls.install_packages(StrVector(pkgs))

    def fit_transform(self, X_train, y_train, X_test):
        """Remove bias from the given dataset by fair adaptation.

        Args:
            X_train (pandas.DataFrame): Training data frame (including the
                protected attribute).
            y_train (pandas.Series): Training labels.
            X_test (pandas.DataFrame): Test data frame (including the protected
                attribute).

        Returns:
            tuple:
                Transformed inputs.

                * **X_fair_train** (pandas.DataFrame) -- Transformed training
                  data.
                * **y_fair_train** (array-like) -- Transformed training labels.
                * **X_fair_test** (pandas.DataFrame) -- Transformed test data.

        """
        # merge X_train and y_train
        df_train = pd.concat([X_train, y_train], axis=1)
        groups, self.prot_attr_ = check_groups(X_train, self.prot_attr, ensure_binary=True)
        self.groups_ = np.unique(groups)

        wrapper = osp.join(osp.dirname(osp.abspath(__file__)), 'fairadapt.R')
        robjects.r.source(wrapper)
        FairAdapt_R = robjects.r['wrapper']
        # convert to Pandas with a local converter
        with localconverter(robjects.default_converter + pandas2ri.converter):
            train_data = robjects.conversion.py2rpy(df_train)
            test_data = robjects.conversion.py2rpy(X_test)
            adj_mat = robjects.conversion.py2rpy(self.adj_mat)

        # run FairAdapt in R
        res = FairAdapt_R(
            train_data=train_data,
            test_data=test_data,
            adj_mat=adj_mat,
            prot_attr=self.prot_attr_,
            outcome=y_train.name
        )

        with localconverter(robjects.default_converter + pandas2ri.converter):
            X_fair_train = robjects.conversion.rpy2py(res.rx2('train'))
            X_fair_test = robjects.conversion.rpy2py(res.rx2('test'))
        X_fair_train.columns = [y_train.name] + X_train.columns.tolist()
        y_fair_train = X_fair_train.pop(y_train.name)
        X_fair_test.columns = X_test.columns

        return X_fair_train, y_fair_train, X_fair_test



from logging import warning

import numpy as np
import pandas as pd

from aif360.datasets import StructuredDataset

from sklearn.preprocessing import MinMaxScaler


class RegressionDataset(StructuredDataset):
    """Base class for regression datasets."""

    def __init__(self, df, dep_var_name, protected_attribute_names,
                 privileged_classes, instance_weights_name='',
                 categorical_features=[], na_values=[],
                 custom_preprocessing=None, metadata=None):
        """
        Subclasses of RegressionDataset should perform the following before
        calling `super().__init__`:

            1. Load the dataframe from a raw file.

        Then, this class will go through a standard preprocessing routine which:

            2. (optional) Performs some dataset-specific preprocessing (e.g.
               renaming columns/values, handling missing data).

            3. Drops rows with NA values.

            4. Creates a one-hot encoding of the categorical variables.

            5. Maps protected attributes to binary privileged/unprivileged
               values (1/0).

            6. Normalizes df values

        Args:
            df (pandas.DataFrame): DataFrame on which to perform standard
                processing.
            dep_var_name: Name of the dependent variable column in `df`.
            protected_attribute_names (list): List of names corresponding to
                protected attribute columns in `df`.
            privileged_classes (list(list or function)): Each element is
                a list of values which are considered privileged or a boolean
                function which return `True` if privileged for the corresponding
                column in `protected_attribute_names`. All others are
                unprivileged. Values are mapped to 1 (privileged) and 0
                (unprivileged) if they are not already numerical.
            instance_weights_name (optional): Name of the instance weights
                column in `df`.
            categorical_features (optional, list): List of column names in the
                DataFrame which are to be expanded into one-hot vectors.
            na_values (optional): Additional strings to recognize as NA. See
                :func:`pandas.read_csv` for details.
            custom_preprocessing (function): A function object which
                acts on and returns a DataFrame (f: DataFrame -> DataFrame). If
                `None`, no extra preprocessing is applied.
            metadata (optional): Additional metadata to append.
        """
        # 2. Perform dataset-specific preprocessing
        if custom_preprocessing:
            df = custom_preprocessing(df)

        # 3. Remove any rows that have missing data.
        dropped = df.dropna()
        count = df.shape[0] - dropped.shape[0]
        if count > 0:
            warning("Missing Data: {} rows removed from {}.".format(count,
                    type(self).__name__))
        df = dropped

        # 4. Create a one-hot encoding of the categorical variables.
        df = pd.get_dummies(df, columns=categorical_features, prefix_sep='=')

        # 5. Map protected attributes to privileged/unprivileged
        privileged_protected_attributes = []
        unprivileged_protected_attributes = []
        for attr, vals in zip(protected_attribute_names, privileged_classes):
            privileged_values = [1.]
            unprivileged_values = [0.]
            if callable(vals):
                df[attr] = df[attr].apply(vals)
            elif np.issubdtype(df[attr].dtype, np.number):
                # this attribute is numeric; no remapping needed
                privileged_values = vals
                unprivileged_values = list(set(df[attr]).difference(vals))
            else:
                # find all instances which match any of the attribute values
                priv = np.logical_or.reduce(np.equal.outer(vals, df[attr].to_numpy()))
                df.loc[priv, attr] = privileged_values[0]
                df.loc[~priv, attr] = unprivileged_values[0]

            privileged_protected_attributes.append(
                np.array(privileged_values, dtype=np.float64))
            unprivileged_protected_attributes.append(
                np.array(unprivileged_values, dtype=np.float64))

        # 6. Normalize df values
        df = pd.DataFrame(MinMaxScaler().fit_transform(df.values),
                          columns=list(df), index=df.index)

        super(RegressionDataset, self).__init__(df=df,
            label_names=[dep_var_name],
            protected_attribute_names=protected_attribute_names,
            privileged_protected_attributes=privileged_protected_attributes,
            unprivileged_protected_attributes=unprivileged_protected_attributes,
            instance_weights_name=instance_weights_name,
            scores_names=[],
            metadata=metadata)



from typing import Union

from aif360.datasets import BinaryLabelDataset
from aif360.metrics import ClassificationMetric

from aif360.detectors.mdss.ScoringFunctions import Bernoulli, BerkJones, ScoringFunction
from aif360.detectors.mdss.MDSS import MDSS

import pandas as pd


class MDSSClassificationMetric(ClassificationMetric):
    """Bias subset scanning is proposed as a technique to identify bias in
    predictive models using subset scanning [#zhang16]_.

    This class is a wrapper for the bias scan scoring and scanning methods that
    uses the ClassificationMetric abstraction.

    References:
        .. [#zhang16] `Zhang, Z. and Neill, D. B., "Identifying significant
           predictive bias in classifiers," arXiv preprint, 2016.
           <https://arxiv.org/abs/1611.08292>`_
    """

    def __init__(
        self,
        dataset: BinaryLabelDataset,
        classified_dataset: BinaryLabelDataset,
        scoring: Union[str, ScoringFunction] = 'Bernoulli',
        unprivileged_groups: dict = None,
        privileged_groups: dict = None,
        **kwargs
    ):
        """
        Args:
            dataset (BinaryLabelDataset): Dataset containing ground-truth
                labels.
            classified_dataset (BinaryLabelDataset): Dataset containing
                predictions.
            scoring (str or ScoringFunction): One of 'Bernoulli' (parametric), or 'BerkJones' (non-parametric)
                        or subclass of :class:`aif360.metrics.mdss.ScoringFunctions.ScoringFunction`.
                        Defaults to Bernoulli.
            privileged_groups (list(dict)): Privileged groups. Format is a list
                of `dicts` where the keys are `protected_attribute_names` and
                the values are values in `protected_attributes`. Each `dict`
                element describes a single group. See examples for more details.
            unprivileged_groups (list(dict)): Unprivileged groups in the same
                format as `privileged_groups`.
        """

        super(MDSSClassificationMetric, self).__init__(
            dataset,
            classified_dataset,
            unprivileged_groups=unprivileged_groups,
            privileged_groups=privileged_groups,
        )

        self.scoring = scoring
        self.kwargs = kwargs

    def score_groups(self, privileged=True, penalty=1e-17):
        """Compute the bias score for a prespecified group of records.

        Args:
            privileged (bool): Flag for which direction to scan: privileged
                (``True``) implies negative (observed worse than predicted
                outcomes) while unprivileged (``False``) implies positive
                (observed better than predicted outcomes).

        Returns:
            float: Bias score for the given group.
                The higher the score, the evidence for bias.
        """
        groups = self.privileged_groups if privileged else self.unprivileged_groups
        subset = dict()

        for g in groups:
            for k, v in g.items():
                if k in subset.keys():
                    subset[k].append(v)
                else:
                    subset[k] = [v]

        coordinates = pd.DataFrame(
            self.dataset.features, columns=self.dataset.feature_names
        )
        expected = pd.Series(self.classified_dataset.scores.flatten())
        outcomes = pd.Series(self.dataset.labels.flatten() == self.dataset.favorable_label, dtype=int)

        # In MDSS, we look for subset whose observations systematically deviates from expectations.
        # Positive direction means observations are systematically higher than expectations
        # (or expectations are systematically higher than observations) while
        # Negative direction means observatons are systematically lower than expectations
        # (or expectations are systematically higher than observations)

        # For a privileged group, we are looking for a subset whose expectations
        # (where expectations is obtained from a model) is systematically higher than the observations.
        # This means we scan in the negative direction.

        # For an uprivileged group, we are looking for a subset whose expectations
        # (where expectations is obtained from a model) is systematically lower the observations.
        # This means we scan in the position direction.

        self.kwargs['direction'] = "negative" if privileged else "positive"

        if self.scoring == "Bernoulli":
            scoring_function = Bernoulli(**self.kwargs)
        elif self.scoring == "BerkJones":
            scoring_function = BerkJones(**self.kwargs)
        else:
            scoring_function = self.scoring(**self.kwargs)

        scanner = MDSS(scoring_function)

        return scanner.score_current_subset(
            coordinates, expected, outcomes, dict(subset), penalty
        )



from aif360.detectors.mdss.ScoringFunctions.ScoringFunction import ScoringFunction
from aif360.detectors.mdss.ScoringFunctions import optim

import numpy as np


class Gaussian(ScoringFunction):
    def __init__(self, **kwargs):
        """
        Gaussian score function. May be appropriate to use when the outcome of
        interest is assumed to be normally distributed.

        kwargs must contain
        'direction (str)' - direction of the severity; could be higher than expected outcomes ('positive') or lower than expected ('negative')
        """

        super(Gaussian, self).__init__(**kwargs)

    def score(
        self, observed_sum: float, expectations: np.array, penalty: float, q: float
    ):
        """
        Computes gaussian bias score for given q

        :param observed_sum: sum of observed binary outcomes for all i
        :param expectations: predicted outcomes for each data element i
        :param penalty: penalty term. Should be positive
        :param q: current value of q
        :return: bias score for the current value of q
        """

        assumed_var =  self.var
        expected_sum = expectations.sum()
        penalty /= self.var

        C = (
            observed_sum * expected_sum / assumed_var * (q - 1)
        ) 

        B = (
            expected_sum**2 * (1 - q**2) / (2 * assumed_var)
        )

        if C > B and self.direction == 'positive':
            ans = C + B
        elif B > C and self.direction == 'negative':
            ans = C + B
        else:
            ans = 0

        ans -= penalty

        return ans

    def qmle(self, observed_sum: float, expectations: np.array):
        """
        Computes the q which maximizes score (q_mle).
        """
        expected_sum = expectations.sum()

        # Deals with case where observed_sum = expected_sum = 0
        if observed_sum == expected_sum:
            ans = 1
        else:
            ans = observed_sum / expected_sum
        
        assert np.isnan(ans) == False, f'{expected_sum}, {observed_sum}, {ans}' 
        return ans

    def compute_qs(self, observed_sum: float, expectations: np.array, penalty: float):
        """
        Computes roots (qmin and qmax) of the score function for given q

        :param observed_sum: sum of observed binary outcomes for all i
        :param expectations: predicted outcomes for each data element i
        :param penalty: penalty coefficient
        """

        direction = self.direction

        q_mle = self.qmle(observed_sum, expectations)
        q_mle_score = self.score(observed_sum, expectations, penalty, q_mle)

        if q_mle_score > 0:
            exist = 1
            q_min = optim.bisection_q_min(self, observed_sum, expectations, penalty, q_mle, temp_min=-1e6)
            q_max = optim.bisection_q_max(self, observed_sum, expectations, penalty, q_mle, temp_max=1e6)
        else:
            # there are no roots
            exist = 0
            q_min = 0
            q_max = 0

        # only consider the desired direction, positive or negative
        if exist:
            exist, q_min, q_max = optim.direction_assertions(direction, q_min, q_max)

        ans = [exist, q_mle, q_min, q_max]
        return ans



import numpy as np

from aif360.datasets import BinaryLabelDataset
from aif360.algorithms import Transformer


class ARTClassifier(Transformer):
    """Wraps an instance of an :obj:`art.classifiers.Classifier` to extend
    :obj:`~aif360.algorithms.Transformer`.
    """

    def __init__(self, art_classifier):
        """Initialize ARTClassifier.

        Args:
            art_classifier (art.classifier.Classifier): A Classifier
                object from the `adversarial-robustness-toolbox`_.

        .. _adversarial-robustness-toolbox:
           https://github.com/Trusted-AI/adversarial-robustness-toolbox
        """
        super(ARTClassifier, self).__init__(art_classifier=art_classifier)
        self._art_classifier = art_classifier

    def fit(self, dataset, batch_size=128, nb_epochs=20):
        """Train a classifer on the input.

        Args:
            dataset (Dataset): Training dataset.
            batch_size (int): Size of batches (passed through to ART).
            nb_epochs (int): Number of epochs to use for training (passed
                through to ART).

        Returns:
            ARTClassifier: Returns self.
        """
        self._art_classifier.fit(dataset.features, dataset.labels,
                                 batch_size=batch_size, nb_epochs=nb_epochs)
        return self

    def predict(self, dataset, logits=False):
        """Perform prediction for the input.

        Args:
            dataset (Dataset): Test dataset.
            logits (bool, optional): True is prediction should be done at the
                logits layer (passed through to ART).

        Returns:
            Dataset: Dataset with predicted labels in the `labels` field.
        """
        pred_labels = self._art_classifier.predict(dataset.features,
            dataset.labels, logits=logits)

        if isinstance(dataset, BinaryLabelDataset):
            pred_labels = np.argmax(pred_labels, axis=1).reshape((-1, 1))

        pred_dataset = dataset.copy()
        pred_dataset.labels = pred_labels

        return pred_dataset



from typing import Union

from aif360.detectors import bias_scan as _bias_scan
from aif360.detectors.mdss.ScoringFunctions import ScoringFunction

import pandas as pd


def bias_scan(
    X: pd.DataFrame,
    y_true: pd.Series,
    y_pred: Union[pd.Series, pd.DataFrame] = None,
    pos_label: Union[str, float] = None,
    overpredicted: bool = True,
    scoring: Union[str, ScoringFunction] = "Bernoulli",
    num_iters: int = 10,
    penalty: float = 1e-17,
    mode: str = "binary",
    **kwargs,
):
    """Scan to find the highest scoring subset of records.

    Args:
        X (pandas.DataFrame): the dataset (containing the features) the model was trained on
        y_true (pandas.Series): ground truth (correct) target values
        y_pred (pandas.Series,  pandas.DataFrame, optional): pandas series estimated targets
            as returned by a model for binary, continuous and ordinal modes.
            If mode is nominal, this is a dataframe with columns containing expectations/predictions for each nominal class.
            If None, model is assumed to be a dumb model that predicts the mean of the targets or 1/(num of categories) for nominal mode.
        pos_label (str, float, optional): Should be high or low or float if the mode in [binary, ordinal, or continuous].
            If float, value has to be minimum or maximum in the y_true column. Defaults to high if None for these modes.
            Support for float left in to keep the intuition clear in binary classification tasks.
            If mode is nominal, favorable values should be one of the unique categories in the y_true column.
            Defaults to a one-vs-all scan if None for nominal mode.
        overpredicted (bool, optional): flag for group to scan for.
            True means we scan for a group whose expectations/predictions are systematically higher than observed.
            In other words, True means we scan for a group whose observeed is systematically lower than the expectations.
            False means we scan for a group whose expectations/predictions are systematically lower than observed.
            In other words, False means we scan for a group whose observed is systematically higher than the expectations.
        scoring (str or class): One of 'Bernoulli', 'Gaussian', 'Poisson', or 'BerkJones' or subclass of
            :class:`aif360.detectors.mdss.ScoringFunctions.ScoringFunction`.
        num_iters (int, optional): number of iterations (random restarts). Should be positive.
        penalty (float,optional): penalty term. Should be positive. The penalty term as with any regularization parameter may need to be
            tuned for ones use case. The higher the penalty, the less complex (number of features and feature values) the
            highest scoring subset that gets returned is.
        mode(str): one of ['binary', 'continuous', 'nominal', 'ordinal']. Defaults to binary.
            In nominal mode, up to 10 categories are supported by default.
            To increase this, pass in keyword argument max_nominal = integer value.

     Returns:
        tuple: The highest scoring subset and the score or dict of the highest scoring subset and the score for each category in nominal mode
    """
    return _bias_scan(
        data=X,
        observations=y_true,
        expectations=y_pred,
        favorable_value=pos_label,
        overpredicted=overpredicted,
        scoring=scoring,
        num_iters=num_iters,
        penalty=penalty,
        mode=mode,
        **kwargs
    )



import numpy as np
import pandas as pd

from aif360.algorithms import Transformer
from aif360.datasets import StructuredDataset, RegressionDataset

class DeterministicReranking(Transformer):
    """A collection of algorithms for construction of fair ranked candidate lists. [1]_ .

    References:
        .. [1] Sahin Cem Geyik, Stuart Ambler, and Krishnaram Kenthapadi,
            "Fairness-Aware Ranking in Search & Recommendation Systems with Application to LinkedIn Talent Search,"
            KDD '19: Proceedings of the 25th ACM SIGKDD International Conference
            on Knowledge Discovery & Data Mining, July 2019, Pages 2221-2231.
    """

    def __init__(self,
            unprivileged_groups,
            privileged_groups):
        """
        Args:
            unprivileged_groups (list(dict)): Representation for the unprivileged
                group.
            privileged_groups (list(dict)): Representation for the privileged
                group.
        """

        super(DeterministicReranking, self).__init__(
            unprivileged_groups=unprivileged_groups,
            privileged_groups=privileged_groups)
        
        self.unprivileged_groups = unprivileged_groups
        self.privileged_groups = privileged_groups
        self._n_groups = len(unprivileged_groups) + len(privileged_groups)
        self.s = set(unprivileged_groups[0].keys())
        self.s_vals = set(self.unprivileged_groups[0].values()).union(set(self.privileged_groups[0].values()))

    def fit(self, dataset: RegressionDataset):
        if list(self.unprivileged_groups[0].keys())[0] != list(self.privileged_groups[0].keys())[0]:
            raise ValueError("Different sensitive attributes (not values) specified for unprivileged and privileged groups.")  

        items = dataset.convert_to_dataframe()[0]
        items = items.sort_values(axis=0, by=dataset.label_names[0], ascending=False)

        if not(self.s.issubset(items.columns)):
            raise ValueError(f"The dataset must contain the protected attribute(s): '{self.s}'.")
        
        # if we have just 1 protected attribute
        if not isinstance(self.s, list) and False:
            self._item_groups = [items[items[self.s] == ai] for ai in self.s_vals]
        else:
            self._item_groups = []
            for group in self.unprivileged_groups + self.privileged_groups:
                q = ' & '.join(
                    [f'{s_i}=="{v_i}"' if isinstance(v_i, str) else f'{s_i}=={v_i}' 
                     for s_i, v_i in group.items()]
                     )
                self._item_groups.append(items.query(q))

        # self._item_groups = {ai: it for ai, it in zip(
        #     self.s_vals, [items[items[self.s] == ai] for ai in self.s_vals])}
        
        return self

    def predict(self,
                    dataset: RegressionDataset,
                    rec_size: int,
                    target_prop: list,
                    rerank_type: str='Constrained',
                    renormalize_scores: bool=False
                    ) -> RegressionDataset:
        """Construct a ranking of candidates in the dataset according to specified proportions of groups.

        Args:
            dataset (RegressionDataset): Dataset to rerank.
            rec_size (int): Number of candidates in the output.
            target_prop (list): Desired proportion of each group in the output.
            rerank_type: Greedy, Conservative, Relaxed, or Constrained. Determines the type of algorithm \
                as described in the original paper.
            renormalize_scores: renormalize label (score) values in the resulting ranking. If True, uses the default \
                behavior of RegressionDataset.

        Returns:
            RegressionDataset: The reranked dataset.
        """
        
        
        if rec_size <= 0:
            raise ValueError(f"Output size should be greater than 0, got {rec_size}.")
        # if np.any(set(target_prop.keys()) != set(self.s_vals)):
        #     raise ValueError("""Proportion specifications do not match. \
        #         `target_prop` should have sensitive attribute values as keys.""")
        if len(dataset.label_names) != 1:
            raise ValueError(f"Dataset must have exactly one label, got {len(dataset.label_names)}.")
        if rerank_type not in ['Greedy', 'Conservative', 'Relaxed', 'Constrained']:
            raise ValueError(f'`rerank_type` must be one of `Greedy`, `Conservative`, `Relaxed`, `Constrained`; got {rerank_type}')
        
        # group_counts = {a: 0 for a in self.s_vals}
        group_counts = [0] * self._n_groups
        rankedItems = []
        score_label = dataset.label_names[0]

        if rerank_type != 'Constrained':
            for k in range(1, rec_size+1):
                below_min, below_max = [], []
                # get the best-scoring candidate item from each group
                candidates = [
                    candidates_gi.iloc[group_counts[g_i]] for g_i, candidates_gi in enumerate(self._item_groups)
                    ]
                for group_idx in range(self._n_groups):
                    # best unranked items for each group
                    if group_counts[group_idx] < np.floor(k*target_prop[group_idx]):
                        below_min.append(group_idx)
                    elif group_counts[group_idx] < np.ceil(k*target_prop[group_idx]):
                        below_max.append(group_idx)
                # if some groups are currently underrepresented
                if len(below_min) != 0:
                    # choose the best next item among currently underrepresented groups
                    candidates_bmin = [candidates[group_idx] for group_idx in below_min]
                    next_group, next_item = max(enumerate(candidates_bmin), key = lambda x: x[1][score_label])
                # if minimal representation requirements are satisfied
                else:
                    # if Greedy, add the highest scoring candidate among the groups
                    if rerank_type == 'Greedy':
                        candidates_bmax = [candidates[group_idx] for group_idx in below_max]
                        next_group, next_item = max(enumerate(candidates_bmax), key = lambda x: x[1][score_label])
                    # if Conservative, add the candidate from the group least represented so far
                    elif rerank_type == 'Conservative':
                        # group_rep = [np.ceil(k*target_prop[group])/target_prop[group] for group in below_max]
                        # sort by how close the groups are to violating the condition, in case of tie sort by best element score
                        next_group = min(below_max, key=lambda group_idx:
                                        (np.ceil(k*target_prop[group_idx])/target_prop[group_idx],
                                         -candidates[group_idx][score_label]))
                        next_item = candidates[next_group]
                    # if Relaxed, relax the conservative requirements
                    elif rerank_type == 'Relaxed':
                        next_group = min(below_max, key=lambda group_idx:
                                            (np.ceil(np.ceil(k*target_prop[group_idx])/target_prop[group_idx]),
                                             -candidates[group_idx][score_label])
                                             )
                        next_item = candidates[next_group]

                rankedItems.append(next_item)
                group_counts[next_group] += 1

        elif rerank_type == 'Constrained':
            rankedItems, maxIndices = [], []
            group_counts, min_counts = [0] * self._n_groups, [0] * self._n_groups

            lastEmpty, k = 0, 0
            while lastEmpty < rec_size:
                k+=1
                # determine the minimum feasible counts of each group at current rec. list size
                min_counts_at_k = [np.floor(p_gi*k) for p_gi in target_prop]
                # get sensitive attr. values for which the current minimum count has increased
                # since last one
                changed_mins = []
                for group_idx in range(self._n_groups):
                    if min_counts_at_k[group_idx] > min_counts[group_idx]:
                        changed_mins.append(group_idx)
                    
                if len(changed_mins) > 0:
                    # get the list of candidates to insert and sort them by their score
                    changed_items = []
                    # save the candidate AND the index of the group it belongs to
                    for group_idx in changed_mins:
                        changed_items.append((group_idx, self._item_groups[group_idx].iloc[group_counts[group_idx]]))
                    changed_items.sort(key=lambda x: -x[1][score_label])

                    # add the candidate items, starting with the best score
                    for newitem in changed_items:
                        if len(rankedItems) == rec_size:
                            break
                        maxIndices.append(k-1)
                        rankedItems.append(newitem[1])
                        swapInd = lastEmpty
                        while swapInd > 0 and maxIndices[swapInd-1] >= swapInd and rankedItems[swapInd-1][score_label] < rankedItems[swapInd][score_label]:
                            maxIndices[swapInd-1], maxIndices[swapInd] = maxIndices[swapInd], maxIndices[swapInd-1]
                            rankedItems[swapInd-1], rankedItems[swapInd] = rankedItems[swapInd], rankedItems[swapInd-1]
                            swapInd -= 1
                        lastEmpty+=1
                        group_counts[newitem[0]] += 1
                    min_counts = min_counts_at_k

        res_df = pd.DataFrame(rankedItems, columns=dataset.feature_names + [score_label])
        res = RegressionDataset(res_df,
                                dep_var_name=dataset.label_names[0],
                                protected_attribute_names=dataset.protected_attribute_names,
                                privileged_classes=dataset.privileged_protected_attributes)
        if not renormalize_scores:
            res.labels = np.transpose([res_df[score_label]])
            res.scores = np.transpose([res_df[score_label]])
        return res
    

    def fit_predict(self,
                    dataset: RegressionDataset,
                    rec_size: int,
                    target_prop: dict,
                    rerank_type: str='Constrained',
                    renormalize_scores: bool=False
                    ) -> RegressionDataset:
        self.fit(dataset=dataset)
        return self.predict(dataset=dataset, 
                            rec_size=rec_size, 
                            target_prop=target_prop, 
                            rerank_type=rerank_type, 
                            renormalize_scores=renormalize_scores)


from aif360.algorithms.preprocessing.disparate_impact_remover import DisparateImpactRemover
from aif360.algorithms.preprocessing.lfr import LFR
from aif360.algorithms.preprocessing.optim_preproc import OptimPreproc
from aif360.algorithms.preprocessing.reweighing import Reweighing



# Original work Copyright (c) 2017 Geoff Pleiss
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Modified work Copyright 2018 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import numpy as np

from aif360.algorithms import Transformer
from aif360.metrics import ClassificationMetric, utils


class CalibratedEqOddsPostprocessing(Transformer):
    """Calibrated equalized odds postprocessing is a post-processing technique
    that optimizes over calibrated classifier score outputs to find
    probabilities with which to change output labels with an equalized odds
    objective [7]_.

    References:
        .. [7] G. Pleiss, M. Raghavan, F. Wu, J. Kleinberg, and
           K. Q. Weinberger, "On Fairness and Calibration," Conference on Neural
           Information Processing Systems, 2017

    Adapted from:
    https://github.com/gpleiss/equalized_odds_and_calibration/blob/master/calib_eq_odds.py
    """

    def __init__(self, unprivileged_groups, privileged_groups,
                 cost_constraint='weighted', seed=None):
        """
        Args:
            unprivileged_groups (dict or list(dict)): Representation for
                unprivileged group.
            privileged_groups (dict or list(dict)): Representation for
                privileged group.
            cost_contraint: fpr, fnr or weighted
            seed (int, optional): Seed to make `predict` repeatable.
        """
        super(CalibratedEqOddsPostprocessing, self).__init__(
            unprivileged_groups=unprivileged_groups,
            privileged_groups=privileged_groups,
            seed=seed)

        self.seed = seed
        self.model_params = None
        self.unprivileged_groups = [unprivileged_groups] \
            if isinstance(unprivileged_groups, dict) else unprivileged_groups
        self.privileged_groups = [privileged_groups] \
            if isinstance(privileged_groups, dict) else privileged_groups
        self.cost_constraint = cost_constraint
        if self.cost_constraint == 'fnr':
            self.fn_rate = 1
            self.fp_rate = 0
        elif self.cost_constraint == 'fpr':
            self.fn_rate = 0
            self.fp_rate = 1
        elif self.cost_constraint == 'weighted':
            self.fn_rate = 1
            self.fp_rate = 1

        self.base_rate_priv = 0.0
        self.base_rate_unpriv = 0.0

    def fit(self, dataset_true, dataset_pred):
        """Compute parameters for equalizing generalized odds using true and
        predicted scores, while preserving calibration.

        Args:
            dataset_true (BinaryLabelDataset): Dataset containing true `labels`.
            dataset_pred (BinaryLabelDataset): Dataset containing predicted
                `scores`.

        Returns:
            CalibratedEqOddsPostprocessing: Returns self.
        """

        # Create boolean conditioning vectors for protected groups
        cond_vec_priv = utils.compute_boolean_conditioning_vector(
            dataset_pred.protected_attributes,
            dataset_pred.protected_attribute_names,
            self.privileged_groups)
        cond_vec_unpriv = utils.compute_boolean_conditioning_vector(
            dataset_pred.protected_attributes,
            dataset_pred.protected_attribute_names,
            self.unprivileged_groups)

        cm = ClassificationMetric(dataset_true, dataset_pred,
                                  unprivileged_groups=self.unprivileged_groups,
                                  privileged_groups=self.privileged_groups)
        self.base_rate_priv = cm.base_rate(privileged=True)
        self.base_rate_unpriv = cm.base_rate(privileged=False)

        # Create a dataset with "trivial" predictions
        dataset_trivial = dataset_pred.copy(deepcopy=True)
        dataset_trivial.scores[cond_vec_priv] = cm.base_rate(privileged=True)
        dataset_trivial.scores[cond_vec_unpriv] = cm.base_rate(privileged=False)
        cm_triv = ClassificationMetric(dataset_true, dataset_trivial,
            unprivileged_groups=self.unprivileged_groups,
            privileged_groups=self.privileged_groups)

        if self.fn_rate == 0:
            priv_cost = cm.generalized_false_positive_rate(privileged=True)
            unpriv_cost = cm.generalized_false_positive_rate(privileged=False)
            priv_trivial_cost = cm_triv.generalized_false_positive_rate(privileged=True)
            unpriv_trivial_cost = cm_triv.generalized_false_positive_rate(privileged=False)

        elif self.fp_rate == 0:
            priv_cost = cm.generalized_false_negative_rate(privileged=True)
            unpriv_cost = cm.generalized_false_negative_rate(privileged=False)
            priv_trivial_cost = cm_triv.generalized_false_negative_rate(privileged=True)
            unpriv_trivial_cost = cm_triv.generalized_false_negative_rate(privileged=False)

        else:
            priv_cost = weighted_cost(self.fp_rate, self.fn_rate, cm, privileged=True)
            unpriv_cost = weighted_cost(self.fp_rate, self.fn_rate, cm, privileged=False)
            priv_trivial_cost = weighted_cost(self.fp_rate, self.fn_rate, cm_triv, privileged=True)
            unpriv_trivial_cost = weighted_cost(self.fp_rate, self.fn_rate, cm_triv, privileged=False)

        unpriv_costs_more = unpriv_cost > priv_cost
        self.priv_mix_rate = (unpriv_cost - priv_cost) / (priv_trivial_cost - priv_cost) if unpriv_costs_more else 0
        self.unpriv_mix_rate = 0 if unpriv_costs_more else (priv_cost - unpriv_cost) / (unpriv_trivial_cost - unpriv_cost)

        return self

    def predict(self, dataset, threshold=0.5):
        """Perturb the predicted scores to obtain new labels that satisfy
        equalized odds constraints, while preserving calibration.

        Args:
            dataset (BinaryLabelDataset): Dataset containing `scores` that needs
                to be transformed.
            threshold (float): Threshold for converting `scores` to `labels`.
                Values greater than or equal to this threshold are predicted to
                be the `favorable_label`. Default is 0.5.
        Returns:
            dataset (BinaryLabelDataset): transformed dataset.
        """
        if self.seed is not None:
            np.random.seed(self.seed)

        cond_vec_priv = utils.compute_boolean_conditioning_vector(
            dataset.protected_attributes,
            dataset.protected_attribute_names,
            self.privileged_groups)
        cond_vec_unpriv = utils.compute_boolean_conditioning_vector(
            dataset.protected_attributes,
            dataset.protected_attribute_names,
            self.unprivileged_groups)

        unpriv_indices = (np.random.random(sum(cond_vec_unpriv))
                       <= self.unpriv_mix_rate)
        unpriv_new_pred = dataset.scores[cond_vec_unpriv].copy()
        unpriv_new_pred[unpriv_indices] = self.base_rate_unpriv

        priv_indices = (np.random.random(sum(cond_vec_priv))
                     <= self.priv_mix_rate)
        priv_new_pred = dataset.scores[cond_vec_priv].copy()
        priv_new_pred[priv_indices] = self.base_rate_priv

        dataset_new = dataset.copy(deepcopy=True)

        dataset_new.scores = np.zeros_like(dataset.scores, dtype=np.float64)
        dataset_new.scores[cond_vec_priv] = priv_new_pred
        dataset_new.scores[cond_vec_unpriv] = unpriv_new_pred

        # Create labels from scores using a default threshold
        dataset_new.labels = np.where(dataset_new.scores >= threshold,
                                      dataset_new.favorable_label,
                                      dataset_new.unfavorable_label)
        return dataset_new

    def fit_predict(self, dataset_true, dataset_pred, threshold=0.5):
        """fit and predict methods sequentially."""
        return self.fit(dataset_true, dataset_pred).predict(
            dataset_pred, threshold=threshold)

######### SUPPORTING FUNCTIONS ##########

def weighted_cost(fp_rate, fn_rate, cm, privileged):
    norm_const = float(fp_rate + fn_rate) if\
                      (fp_rate != 0 and fn_rate != 0) else 1
    return ((fp_rate / norm_const
            * cm.generalized_false_positive_rate(privileged=privileged)
            * (1 - cm.base_rate(privileged=privileged))) +
           (fn_rate / norm_const
            * cm.generalized_false_negative_rate(privileged=privileged)
            * cm.base_rate(privileged=privileged)))



# Metrics function
from collections import OrderedDict
from aif360.metrics import ClassificationMetric

def compute_metrics(dataset_true, dataset_pred, 
                    unprivileged_groups, privileged_groups,
                    disp = True):
    """ Compute the key metrics """
    classified_metric_pred = ClassificationMetric(dataset_true,
                                                 dataset_pred, 
                                                 unprivileged_groups=unprivileged_groups,
                                                 privileged_groups=privileged_groups)
    metrics = OrderedDict()
    metrics["Balanced accuracy"] = 0.5*(classified_metric_pred.true_positive_rate()+
                                             classified_metric_pred.true_negative_rate())
    metrics["Statistical parity difference"] = classified_metric_pred.statistical_parity_difference()
    metrics["Disparate impact"] = classified_metric_pred.disparate_impact()
    metrics["Average odds difference"] = classified_metric_pred.average_odds_difference()
    metrics["Equal opportunity difference"] = classified_metric_pred.equal_opportunity_difference()
    metrics["Theil index"] = classified_metric_pred.theil_index()
    
    if disp:
        for k in metrics:
            print("%s = %.4f" % (k, metrics[k]))
    
    return metrics


import numpy as np
from aif360.metrics import DatasetMetric
from aif360.datasets import RegressionDataset


class RegressionDatasetMetric(DatasetMetric):
    """Class for computing metrics based on a single
    :obj:`~aif360.datasets.RegressionDataset`.
    """

    def __init__(self, dataset, unprivileged_groups=None, privileged_groups=None):
        """
        Args:
            dataset (RegressionDataset): A RegressionDataset.
            privileged_groups (list(dict)): Privileged groups. Format is a list
                of `dicts` where the keys are `protected_attribute_names` and
                the values are values in `protected_attributes`. Each `dict`
                element describes a single group. See examples for more details.
            unprivileged_groups (list(dict)): Unprivileged groups in the same
                format as `privileged_groups`.

        Raises:
            TypeError: `dataset` must be a
                :obj:`~aif360.datasets.RegressionDataset` type.
        """
        if not isinstance(dataset, RegressionDataset):
            raise TypeError("'dataset' should be a RegressionDataset")

        # sets self.dataset, self.unprivileged_groups, self.privileged_groups
        super(RegressionDatasetMetric, self).__init__(dataset,
            unprivileged_groups=unprivileged_groups,
            privileged_groups=privileged_groups)
        
    def infeasible_index(self, target_prop: dict, r: int = None):
        """
        Infeasible Index metric, as described in [1]_.

        Args:
            target_prop (dict): desired proportion of groups.
            r (int): size of the candidate list over which the metric is calculated.
            Defaults to the size of the dataset.
        
        Returns:
            A tuple (int, set{int}): InfeasibleIndex and the positions at which the 
            feasibility condition is violated. 
        
    References:
        .. [1] Sahin Cem Geyik, Stuart Ambler, and Krishnaram Kenthapadi,
            "Fairness-Aware Ranking in Search & Recommendation Systems with Application to LinkedIn Talent Search,"
            KDD '19: Proceedings of the 25th ACM SIGKDD International Conference
            on Knowledge Discovery & Data Mining, July 2019, Pages 2221-2231.
        """
        pr_attr_values = np.ravel(
            self.dataset.unprivileged_protected_attributes + self.dataset.privileged_protected_attributes)
        if set(list(target_prop.keys())) != set(pr_attr_values):
            raise ValueError('Desired proportions must be specified for all values of the protected attributes!')
        
        ranking = np.column_stack((self.dataset.scores, self.dataset.protected_attributes))
        if r is None:
            r = np.ravel(self.dataset.scores).shape[0]
        ii = 0
        k_viol = set()
        for k in range(1, r):
            rk = ranking[:k]
            for ai in pr_attr_values:
                count_ai = rk[rk[:,1] == ai].shape[0]
                if count_ai < np.floor(target_prop[ai]*k):
                    ii+=1
                    k_viol.add(k-1)
        return ii, list(k_viol)
    
    def discounted_cum_gain(self, r: int = None, full_dataset: RegressionDataset=None, normalized=False):
        """
        Discounted Cumulative Gain metric.

        Args:
            r (int): position up to which to calculate the DCG. If not specified, is set to the size of the dataset.
            normalized (bool): return normalized DCG.
            
        Returns:
            The calculated DCG.
        """
        if r is None:
            r = np.ravel(self.dataset.scores).shape[0]
        if r < 0:
            raise ValueError(f'r must be >= 0, got {r}')
        if normalized == True and full_dataset is None:
            raise ValueError('`normalized` is set to True, but `full_dataset` is not specified')
        if not isinstance(full_dataset, RegressionDataset) and not (full_dataset is None):
            raise TypeError(f'`full_datset`: expected `RegressionDataset`, got {type(full_dataset)}')
        scores = np.ravel(self.dataset.scores)[:r]
        z = self._dcg(scores)
        if normalized:
            z /= self._dcg(np.sort(np.ravel(full_dataset.scores))[::-1][:r])
        return z
    
    def _dcg(self, scores):
        logs = np.log2(np.arange(2, len(scores)+2))
        z = np.sum(scores/logs)
        return z



from aif360.detectors.mdss.ScoringFunctions.ScoringFunction import ScoringFunction
from aif360.detectors.mdss.generator import get_entire_subset, get_random_subset


from aif360.detectors.mdss.ScoringFunctions import (
    Bernoulli,
    BerkJones,
    Gaussian,
    ScoringFunction,
    Poisson,
)

import pandas as pd
import numpy as np


class MDSS(object):

    def __init__(self, scoring_function: ScoringFunction):
        self.scoring_function = scoring_function

    def get_aggregates(self, coordinates: pd.DataFrame, outcomes: pd.Series, expectations: pd.Series,
                       current_subset: dict, column_name: str, penalty: float):
        """
        Conditioned on the current subsets of values for all other attributes,
        compute the summed outcome (observed_sum = \sum_i y_i) and all expectations p_i
        for each value of the current attribute.
        Also use additive linear-time subset scanning to compute the set of distinct thresholds
        for which different subsets of attribute values have positive scores. Note that the number
        of such thresholds will be linear rather than exponential in the arity of the attribute.

        :param coordinates: data frame containing having as columns the covariates/features
        :param expectations: data series containing the expectations/expected outcomes
        :param outcomes: data series containing the outcomes/observed outcomes
        :param current_subset: current subset to compute aggregates
        :param column_name: attribute name to scan over
        :param penalty: penalty coefficient
        :return: dictionary of aggregates, sorted thresholds (roots), observed sum of the subset, array of observed
        expectations
        """

        # compute the subset of records matching the current subgroup along all other dimensions
        # temp_df includes the covariates x_i, outcome y_i, and predicted expectation p_i for each matching record
        if current_subset:
            to_choose = coordinates[current_subset.keys()].isin(current_subset).all(axis=1)
            temp_df = pd.concat([coordinates.loc[to_choose], outcomes[to_choose], expectations[to_choose]], axis=1)
        else:
            temp_df = pd.concat([coordinates, outcomes, expectations], axis=1)

        # these wil be used to keep track of the aggregate values and the distinct thresholds to be considered
        aggregates = {}
        thresholds = set()

        scoring_function = self.scoring_function

        # consider each distinct value of the given attribute (column_name)
        for name, group in temp_df.groupby(column_name):
            # compute the sum of outcomes \sum_i y_i
            observed_sum = group.iloc[:, -2].sum()

            # all expectations p_i
            expectations = group.iloc[:, -1].values

            # compute q_min and q_max for the attribute value
            exist, q_mle, q_min, q_max = scoring_function.compute_qs(observed_sum, expectations, penalty)

            # Add to aggregates, and add q_min and q_max to thresholds.
            # Note that thresholds is a set so duplicates will be removed automatically.
            if exist:
                aggregates[name] = {
                    'q_mle': q_mle,
                    'q_min': q_min,
                    'q_max': q_max,
                    'observed_sum': observed_sum,
                    'expectations': expectations
                }
                thresholds.update([q_min, q_max])

        # We also keep track of the summed outcomes \sum_i y_i and the expectations p_i for the case where _
        # all_ values of that attribute are considered (regardless of whether they contribute positively to score).
        # This is necessary because of the way we compute the penalty term: including all attribute values, equivalent
        # to ignoring the attribute, has the lowest penalty (of 0) and thus we need to score that subset as well.
        all_observed_sum = temp_df.iloc[:, -2].sum()
        all_expectations = temp_df.iloc[:, -1].values

        return [aggregates, sorted(thresholds), all_observed_sum, all_expectations]

    def choose_aggregates(self, aggregates: dict, thresholds: list, penalty: float, all_observed_sum: float,
                          all_expectations: list):
        """
        Having previously computed the aggregates and the distinct q thresholds
        to consider in the get_aggregates function,we are now ready to choose the best
        subset of attribute values for the given attribute.
        For each range defined by these thresholds, we will choose all of the positive contributions,
        compute the MLE value of q, and the corresponding score.
        We then pick the best q and score over all of the ranges considered.

        :param aggregates: dictionary of aggregates. For each feature value, it has q_mle, q_min, q_max, observed_sum,
        and the expectations
        :param thresholds: sorted thresholds (roots)
        :param penalty: penalty coefficient
        :param all_observed_sum: sum of observed binary outcomes for all i
        :param all_expectations: data series containing all the expectations/expected outcomes
        :return:
        """
        # initialize
        best_score = 0
        best_names = []

        scoring_function = self.scoring_function

        # for each threshold
        for i in range(len(thresholds) - 1):
            threshold = (thresholds[i] + thresholds[i + 1]) / 2
            observed_sum = 0.0
            expectations = []
            names = []

            # keep only the aggregates which have a positive contribution to the score in that q range
            # we must keep track of the sum of outcome values as well as all predicted expectations
            for key, value in aggregates.items():
                if (value['q_min'] < threshold) & (value['q_max'] > threshold):
                    names.append(key)
                    observed_sum += value['observed_sum']
                    expectations = expectations + value['expectations'].tolist()

            if len(expectations) == 0:
                continue

            # compute the MLE value of q, making sure to only consider the desired direction (positive or negative)
            expectations = np.asarray(expectations)
            current_q_mle = scoring_function.qmle(observed_sum, expectations)

            # Compute the score for the given subset at the MLE value of q.
            # Notice that each included value gets a penalty, so the total penalty
            # is multiplied by the number of included values.
            current_interval_score = scoring_function.score(observed_sum, expectations, penalty * len(names), current_q_mle)

            # keep track of the best score, best q, and best subset of attribute values found so far
            if current_interval_score > best_score:
                best_score = current_interval_score
                best_names = names

        # Now we also have to consider the case of including all attribute values,
        # including those that never make positive contributions to the score.
        # Note that the penalty term is 0 in this case.  (We are neglecting penalties
        # from all other attributes, just considering the current attribute.)

        # compute the MLE value of q, making sure to only consider the desired direction (positive or negative)
        current_q_mle = scoring_function.qmle(all_observed_sum, all_expectations)

        # Compute the score for the given subset at the MLE value of q.
        # Again, the penalty (for that attribute) is 0 when all attribute values are included.
        
        current_score = scoring_function.score(all_observed_sum, all_expectations, 0, current_q_mle)

        # Keep track of the best score, best q, and best subset of attribute values found.
        # Note that if the best subset contains all values of the given attribute,
        # we return an empty list for best_names.
        if current_score > best_score:
            best_score = current_score
            best_names = []

        return [best_names, best_score]

    def score_current_subset(self, coordinates: pd.DataFrame, expectations: pd.Series, outcomes: pd.Series,
                             current_subset: dict, penalty: float):
        """
        Just scores the subset without performing ALTSS.
        We still need to determine the MLE value of q.

        :param coordinates: data frame containing having as columns the covariates/features
        :param expectations: data series containing the expectations/expected outcomes
        :param outcomes: data series containing the outcomes/observed outcomes
        :param current_subset: current subset to be scored
        :param penalty: penalty coefficient
        :return: penalized score of subset
        """

        # compute the subset of records matching the current subgroup along all dimensions
        # temp_df includes the covariates x_i, outcome y_i, and predicted expectation p_i for each matching record
        if current_subset:
            to_choose = coordinates[current_subset.keys()].isin(current_subset).all(axis=1)
            temp_df = pd.concat([coordinates.loc[to_choose], outcomes[to_choose], expectations[to_choose]], axis=1)
        else:
            temp_df = pd.concat([coordinates, outcomes, expectations], axis=1)

        scoring_function = self.scoring_function

        # we must keep track of the sum of outcome values as well as all predicted expectations
        observed_sum = temp_df.iloc[:, -2].sum()
        expectations = temp_df.iloc[:, -1].values

        # compute the MLE value of q, making sure to only consider the desired direction (positive or negative)
        current_q_mle = scoring_function.qmle(observed_sum, expectations)

        # total_penalty = penalty * sum of list lengths in current_subset
        total_penalty = 0
        for key, values in current_subset.items():
            total_penalty += len(values)

        total_penalty *= penalty

        # Compute and return the penalized score    
        penalized_score = scoring_function.score(observed_sum, expectations, total_penalty, current_q_mle)
        return np.round(penalized_score, 4)

    def scan(self, coordinates: pd.DataFrame, expectations: pd.Series, outcomes: pd.Series, penalty: float,
                    num_iters: int, verbose: bool = False, seed: int = 0, mode: str = 'binary'):
        """
        :param coordinates: data frame containing having as columns the covariates/features
        :param expectations: data series containing the expectations/expected outcomes
        :param outcomes: data series containing the outcomes/observed outcomes
        :param penalty: penalty coefficient
        :param num_iters: number of iteration
        :param verbose: logging flag
        :param seed: numpy seed. Default equals 0
        :param mode: one of ['binary', 'continuous', 'nominal', 'ordinal']. Defaults to binary.
        :return: [best subset, best score]
        """
        np.random.seed(seed)

        # Reset indexes
        coordinates = coordinates.reset_index(drop = True)
        expectations = expectations.reset_index(drop = True)
        outcomes = outcomes.reset_index(drop = True)

        assert len(coordinates) == len(expectations) == len(outcomes), \
            f'Lengths of coordinates, expectations, and outcomes should be equal.'

        # Check that the appropriate scoring function is used

        if isinstance(self.scoring_function, BerkJones):
            modes = ["binary", "continuous", "nominal", "ordinal"]
            assert mode in modes, f"Expected one of {modes} for BerkJones,  got {mode}."

            # Ensure that BerkJones only work in Autostrat mode
            unique_expectations = expectations.unique()
            if isinstance(self.scoring_function, BerkJones) and len(unique_expectations) != 1:
                raise Exception(
                    "BerkJones scorer supports scanning in autostrat mode only."
                )

            # Bin the continuous outcomes column for Berk Jones in continuous mode
            alpha = self.scoring_function.alpha
            direction = self.scoring_function.direction
            
            if mode == "continuous":
                quantile = outcomes.quantile(alpha)
                outcomes = (outcomes > quantile).apply(int)

            # Flip outcomes to scan in the negative direction for BerkJones
            # This is equivalent to switching the p-values
            if direction == "negative":
                outcomes = 1 - outcomes

        if isinstance(self.scoring_function, Bernoulli):
            modes = ["binary", "nominal"]
            assert mode in modes, f"Expected one of {modes} for Bernoulli,  got {mode}."

        if isinstance(self.scoring_function, Gaussian):
            assert mode == 'continuous', f"Expected continuous, got {mode}."

             # Set variance for Gaussian
            self.scoring_function.var = expectations.var()
            
            # Move entire distribution to the positive axis
            shift = np.abs(expectations.min()) + np.abs(outcomes.min())
            outcomes = outcomes + shift
            expectations = expectations + shift

        if isinstance(self.scoring_function, Poisson):
            modes = ["binary", "ordinal"]
            assert mode in modes, f"Expected one of {modes} for Poisson,  got {mode}."

        # initialize
        best_subset = {}
        best_score = -1e10
        best_scores = []
        for i in range(num_iters):
            # flags indicates that the method has optimized over subsets for a given attribute.
            # The iteration ends when it cannot further increase score by optimizing over
            # subsets of any attribute, i.e., when all flags are 1.
            flags = np.empty(len(coordinates.columns))
            flags.fill(0)

            # Starting subset. Note that we start with all values for the first iteration
            # and random values for succeeding iterations.
            current_subset = get_entire_subset() if (i == 0) \
                else get_random_subset(coordinates, np.random.rand(1).item(), 10)

            # score the entire population
            current_score = self.score_current_subset(
                coordinates=coordinates,
                expectations=expectations,
                outcomes=outcomes,
                penalty=penalty,
                current_subset=current_subset
            )

            while flags.sum() < len(coordinates.columns):

                # choose random attribute that we haven't scanned yet
                attribute_number_to_scan = np.random.choice(len(coordinates.columns))
                while flags[attribute_number_to_scan]:
                    attribute_number_to_scan = np.random.choice(len(coordinates.columns))
                attribute_to_scan = coordinates.columns.values[attribute_number_to_scan]

                # clear current subset of attribute values for that subset
                if attribute_to_scan in current_subset:
                    del current_subset[attribute_to_scan]

                # call get_aggregates and choose_aggregates to find best subset of attribute values
                aggregates, thresholds, all_observed_sum, all_expectations = self.get_aggregates(
                    coordinates=coordinates,
                    outcomes=outcomes,
                    expectations=expectations,
                    current_subset=current_subset,
                    column_name=attribute_to_scan,
                    penalty=penalty
                )

                temp_names, temp_score = self.choose_aggregates(
                    aggregates=aggregates,
                    thresholds=thresholds,
                    penalty=penalty,
                    all_observed_sum=all_observed_sum,
                    all_expectations=all_expectations
                )

                temp_subset = current_subset.copy()
                # if temp_names is not empty (or null)
                if temp_names:
                    temp_subset[attribute_to_scan] = temp_names

                # Note that this call to score_current_subset ensures that
                # we are penalizing complexity for all attribute values.
                # The value of temp_score computed by choose_aggregates
                # above includes only the penalty for the current attribute.
                temp_score = self.score_current_subset(
                    coordinates=coordinates,
                    expectations=expectations,
                    outcomes=outcomes,
                    penalty=penalty,
                    current_subset=temp_subset
                )

                # reset flags to 0 if we have improved score
                if temp_score > current_score + 1E-6:
                    flags.fill(0)

                # sanity check to make sure score has not decreased
                # sanity check may not apply to Gaussian in penalized mode (TODO: to check Maths again)
                if not isinstance(self.scoring_function, Gaussian) and penalty > 0:
                    assert (
                        temp_score >= current_score - 1e-6
                    ), "WARNING SCORE HAS DECREASED from %.6f to %.6f" % (
                        current_score,
                        temp_score,
                    )
                    
                flags[attribute_number_to_scan] = 1
                current_subset = temp_subset
                current_score = temp_score

            # print out results for current iteration
            if verbose:
                print("Subset found on iteration", i + 1, "of", num_iters, "with score", current_score, ":")
                print(current_subset)

            # update best_score and best_subset if necessary
            if current_score > best_score:
                best_subset = current_subset.copy()
                best_score = current_score

                if verbose:
                    print("Best score is now", best_score)

            elif verbose:
                print("Current score of", current_score, "does not beat best score of", best_score)
            best_scores.append(best_score)
        return best_subset, best_score



import numpy as np
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import accuracy_score, make_scorer

from aif360.algorithms.preprocessing import Reweighing as OrigReweighing
from aif360.sklearn.preprocessing import Reweighing, ReweighingMeta


def test_reweighing_sex(old_adult, new_adult):
    """Test that the old and new Reweighing produce the same sample_weights."""
    X, y, sample_weight = new_adult
    orig_rew = OrigReweighing(unprivileged_groups=[{'sex': 0}],
                              privileged_groups=[{'sex': 1}])
    adult_fair = orig_rew.fit_transform(old_adult)
    rew = Reweighing('sex')
    _, new_sample_weight = rew.fit_transform(X, y, sample_weight=sample_weight)

    assert np.allclose([[orig_rew.w_up_unfav, orig_rew.w_up_fav],
                        [orig_rew.w_p_unfav, orig_rew.w_p_fav]],
                       rew.reweigh_factors_)
    assert np.allclose(adult_fair.instance_weights, new_sample_weight)

def test_reweighing_intersection(new_adult):
    """Test that the new Reweighing runs with >2 protected groups."""
    X, y, _ = new_adult
    rew = Reweighing()
    rew.fit_transform(X, y)
    assert rew.reweigh_factors_.shape == (4, 2)

def test_gridsearch(new_adult):
    """Test that ReweighingMeta works in a grid search."""
    X, y, sample_weight = new_adult
    rew = ReweighingMeta(LogisticRegression(solver='liblinear'), Reweighing())

    # UGLY workaround for sklearn issue: https://stackoverflow.com/a/49598597
    def score_func(y_true, y_pred, sample_weight):
        return accuracy_score(y_true, y_pred, sample_weight=sample_weight.iloc[y_true.index])
    scoring = make_scorer(score_func, sample_weight=sample_weight)

    params = {'estimator__C': [1, 10], 'reweigher__prot_attr': ['sex']}

    clf = GridSearchCV(rew, params, scoring=scoring, cv=5)
    # need to reset index for score_func to work
    clf.fit(X, y.reset_index(drop=True), sample_weight=sample_weight)



from aif360.detectors.mdss.ScoringFunctions.ScoringFunction import ScoringFunction
from aif360.detectors.mdss.ScoringFunctions import optim

import numpy as np


class BerkJones(ScoringFunction):
    def __init__(self, **kwargs):
        """
        Berk-Jones score function is a non parametric expectatation based
        scan statistic that also satisfies the ALTSS property; Non-parametric scoring functions
        do not make parametric assumptions about the model or outcome [1].

        kwargs must contain
        'direction (str)' - direction of the severity; could be higher than expected outcomes ('positive') or lower than expected ('negative')
        'alpha (float)' - the alpha threshold that will be used to compute the score.
            In practice, it may be useful to search over a grid of alpha thresholds and select the one with the maximum score.


        [1] Neill, D. B., & Lingwall, J. (2007). A nonparametric scan statistic for multivariate disease surveillance. Advances in
        Disease Surveillance, 4(106), 570
        """

        super(BerkJones, self).__init__(**kwargs)
        self.alpha = self.kwargs.get('alpha')
        assert self.alpha is not None, "Warning: calling Berk Jones without alpha"

        if self.direction == 'negative':
            self.alpha = 1 - self.alpha


    def score(self, observed_sum: float, expectations: np.array, penalty: float, q: float):
        """
        Computes berk jones score for given q

        :param observed_sum: sum of observed binary outcomes for all i
        :param expectations: predicted outcomes for each data element i
        :param penalty: penalty term. Should be positive
        :param q: current value of q
        :return: berk jones score for the current value of q
        """
        alpha = self.alpha

        if q < alpha:
            q = alpha

        assert q > 0, (
            "Warning: calling compute_score_given_q with "
            "observed_sum=%.2f, expectations of length=%d, penalty=%.2f, q=%.2f, alpha=%.3f"
            % (observed_sum, len(expectations), penalty, q, alpha)
        )
        if q == 1:
            ans = observed_sum * np.log(q / alpha) - penalty
            return ans

        a = observed_sum * np.log(q / alpha)
        b = (len(expectations) - observed_sum) * np.log((1 - q) / (1 - alpha))
        ans = (
            a
            + b
            - penalty
        )

        return ans

    def qmle(self, observed_sum: float, expectations: np.array):
        """
        Computes the q which maximizes score (q_mle).
        for berk jones this is given to be N_a/N
        :param observed_sum: sum of observed binary outcomes for all i
        :param expectations: predicted outcomes for each data element i
        :param direction: direction not considered
        :return: q MLE
        """
        alpha = self.alpha

        if len(expectations) == 0:
            return 0
        else:
            q = observed_sum / len(expectations)

        if (q < alpha):
            return alpha

        return q

    def compute_qs(self, observed_sum: float, expectations: np.array, penalty: float):
        """
        Computes roots (qmin and qmax) of the score function for given q

        :param observed_sum: sum of observed binary outcomes for all i
        :param expectations: predicted outcomes for each data element i
        :param penalty: penalty coefficient
        """
        alpha = self.alpha
        q_mle = self.qmle(observed_sum, expectations)

        if self.score(observed_sum, expectations, penalty, q_mle) > 0:
            exist = 1
            q_min = optim.bisection_q_min(
                self, observed_sum, expectations, penalty, q_mle, temp_min=alpha
            )
            q_max = optim.bisection_q_max(
                self, observed_sum, expectations, penalty, q_mle, temp_max=1
            )
        else:
            # there are no roots
            exist = 0
            q_min = 0
            q_max = 0

        ans = [exist, q_mle, q_min, q_max]
        return ans



import warnings

import numpy as np
import pandas as pd
import scipy.optimize as optim
from scipy.spatial.distance import cdist
from scipy.special import softmax
from sklearn.base import BaseEstimator, ClassifierMixin, TransformerMixin
from sklearn.exceptions import ConvergenceWarning
from sklearn.preprocessing import LabelEncoder
from sklearn.utils import check_random_state
try:
    import torch
    import torch.nn.functional as F
except ImportError as error:
    from logging import warning
    warning("{}: LearnedFairRepresentations will be unavailable. To install, run:\n"
            "pip install 'aif360[LFR]'".format(error))

from aif360.sklearn.utils import check_inputs, check_groups


class LearnedFairRepresentations(BaseEstimator, ClassifierMixin, TransformerMixin):
    """Learned Fair Representations.

    Learned fair representations is a pre-processing technique that finds a
    latent representation which encodes the data well but obfuscates information
    about protected attributes [#zemel13]_. It can also be used as an in-
    processing method by utilizing the learned target coefficients.

    References:
        .. [#zemel13] `R. Zemel, Y. Wu, K. Swersky, T. Pitassi, and C. Dwork,
           "Learning Fair Representations." International Conference on Machine
           Learning, 2013. <http://proceedings.mlr.press/v28/zemel13.html>`_

    # Based on code from https://github.com/zjelveh/learning-fair-representations

    Attributes:
        prot_attr_ (str or list(str)): Protected attribute(s) used for
            reweighing.
        groups_ (array, shape (n_groups,)): A list of group labels known to the
            transformer.
        classes_ (array, shape (n_classes,)): A list of class labels known to
            the transformer.
        priv_group_ (scalar): The label of the privileged group.
        coef_ (array, shape (n_prototypes, 1) or (n_prototypes, n_classes)):
            Coefficient of the intermediate representation for classification.
        prototypes_ (array, shape (n_prototypes, n_features)): The prototype set
            used to form a probabilistic mapping to the intermediate
            representation. These act as clusters and are in the same space as
            the samples.
        n_iter_ (int): Actual number of iterations.
    """

    def __init__(self, prot_attr=None, n_prototypes=5, reconstruct_weight=0.01,
                 target_weight=1., fairness_weight=50., tol=1e-4, max_iter=200,
                 verbose=0, random_state=None):
        """
        Args:
            prot_attr (single label or list-like, optional): Protected
                attribute(s) to use in the reweighing process. If more than one
                attribute, all combinations of values (intersections) are
                considered. Default is ``None`` meaning all protected attributes
                from the dataset are used.
            n_prototypes (int, optional): Size of the set of "prototypes," Z.
            reconstruct_weight (float, optional): Weight coefficient on the L_x
                loss term, A_x.
            target_weight (float, optional): Weight coefficient on the L_y loss
                term, A_y.
            fairness_weight (float, optional): Weight coefficient on the L_z
                loss term, A_z.
            tol (float, optional): Tolerance for stopping criteria.
            max_iter (int, optional): Maximum number of iterations taken for the
                solver to converge.
            verbose (int, optional): Verbosity. 0 = silent, 1 = final loss only,
                2 = print loss every 50 iterations.
            random_state (int or numpy.RandomState, optional): Seed of pseudo-
                random number generator for shuffling data and seeding weights.
        """
        self.prot_attr = prot_attr
        self.n_prototypes = n_prototypes
        self.reconstruct_weight = reconstruct_weight
        self.target_weight = target_weight
        self.fairness_weight = fairness_weight
        self.tol = tol
        self.max_iter = max_iter
        self.verbose = verbose
        self.random_state = random_state

    def fit(self, X, y, priv_group=1, sample_weight=None):
        """Compute the transformation parameters that lead to fair
        representations.

        Args:
            X (pandas.DataFrame): Training samples.
            y (array-like): Training labels.
            priv_group (scalar, optional): The label of the privileged group.
            sample_weight (array-like, optional): Sample weights.

        Returns:
            self
        """
        X, y, sample_weight = check_inputs(X, y, sample_weight)
        rng = check_random_state(self.random_state)

        groups, self.prot_attr_ = check_groups(X, self.prot_attr)
        priv = (groups == priv_group)
        self.priv_group_ = priv_group
        self.groups_ = np.unique(groups)

        le = LabelEncoder()
        y = le.fit_transform(y)
        self.classes_ = le.classes_
        n_classes = len(self.classes_)
        if n_classes == 2:
            n_classes = 1  # XXX
        n_feat = X.shape[1]
        w_size = self.n_prototypes*n_classes

        i = 0
        eps = np.finfo(np.float64).eps

        def LFR_optim_objective(x, X, y, priv):
            nonlocal i
            x = torch.as_tensor(x).requires_grad_()
            w = x[:w_size].view(-1, n_classes)
            v = x[w_size:].view(-1, n_feat)

            M = torch.softmax(-torch.cdist(X, v), dim=1)
            y_pred = M.matmul(w).squeeze(1)

            L_x = F.mse_loss(M.matmul(v), X)
            L_y = F.cross_entropy(y_pred, y) if n_classes > 1 else \
                  F.binary_cross_entropy(y_pred.clamp(eps, 1-eps), y.type_as(w))
            L_z = F.l1_loss(torch.mean(M[priv], 0), torch.mean(M[~priv], 0))
            loss = (self.reconstruct_weight * L_x + self.target_weight * L_y
                  + self.fairness_weight * L_z)

            loss.backward()
            if self.verbose > 1 and i % 50 == 0:
                print("iter: {:{}d}, loss: {:7.3f}, A_x*L_x: {:7.3f}, A_y*L_y: "
                      "{:7.3f}, A_z*L_z: {:7.3f}".format(i,
                        int(np.log10(self.max_iter)+1), loss,
                        self.reconstruct_weight*L_x, self.target_weight*L_y,
                        self.fairness_weight*L_z))
            i += 1
            return loss.item(), x.grad.numpy()

        x0 = rng.random(w_size + self.n_prototypes*n_feat)
        bounds = [(0, 1)]*w_size + [(None, None)]*self.n_prototypes*n_feat
        res = optim.minimize(LFR_optim_objective, x0=x0, method='L-BFGS-B',
                args=(torch.tensor(X.to_numpy(dtype=x0.dtype)), torch.as_tensor(y), priv),
                jac=True, bounds=bounds, options={'gtol': self.tol,
                'maxiter': self.max_iter})

        self.coef_ = res.x[:w_size].reshape(-1, n_classes)
        self.prototypes_ = res.x[w_size:].reshape(-1, n_feat)
        self.n_iter_ = res.nit

        if res.status == 0 and self.verbose:
            print("Converged! iter: {}, loss: {:.3f}".format(res.nit, res.fun))
        elif res.status == 1:
            warnings.warn('lbfgs failed to converge. Increase the number of '
                          'iterations.', ConvergenceWarning)
        elif res.status == 2:
            warnings.warn('lbfgs failed to converge: {}'.format(
                          res.message.decode()), ConvergenceWarning)
        return self

    def transform(self, X):
        """Transform the dataset using the learned model parameters.

        Args:
            X (pandas.DataFrame): Training samples.

        Returns:
            pandas.DataFrame: Transformed samples.
        """
        M = softmax(-cdist(X, self.prototypes_), axis=1)
        Xt = M.dot(self.prototypes_)
        return pd.DataFrame(Xt, columns=X.columns, index=X.index)

    def predict_proba(self, X):
        """Transform the targets using the learned model parameters.

        Args:
            X (pandas.DataFrame): Training samples.

        Returns:
            numpy.ndarray: Transformed targets. Returns the probability of the
            sample for each class in the model, where classes are ordered as
            they are in ``self.classes_``.
        """
        M = softmax(-cdist(X, self.prototypes_), axis=1)
        yt = M.dot(self.coef_)
        if yt.shape[1] == 1:
            yt = np.c_[1-yt, yt]
        else:
            yt = softmax(yt, axis=1)
        return yt

    def predict(self, X):
        """Transform the targets using the learned model parameters.

        Args:
            X (pandas.DataFrame): Training samples.

        Returns:
            numpy.ndarray: Transformed targets.
        """
        probas = self.predict_proba(X)
        return self.classes_[probas.argmax(axis=1)]



"""
The code for GridSearchReduction wraps the source class
fairlearn.reductions.GridSearch
available in the https://github.com/fairlearn/fairlearn library
licensed under the MIT Licencse, Copyright Microsoft Corporation
"""
from logging import warning

try:
    import fairlearn.reductions as red
except ImportError as error:
    warning("{}: GridSearchReduction will be unavailable. To install, run:\n"
            "pip install 'aif360[Reductions]'".format(error))
import pandas as pd

from aif360.algorithms import Transformer
from aif360.sklearn.inprocessing import GridSearchReduction as skGridSearchRed


class GridSearchReduction(Transformer):
    """Grid search reduction for fair classification or regression.

    Grid search is an in-processing technique that can be used for fair
    classification or fair regression. For classification it reduces fair
    classification to a sequence of cost-sensitive classification problems,
    returning the deterministic classifier with the lowest empirical error
    subject to fair classification constraints [#agarwal18]_ among the
    candidates searched. For regression it uses the same priniciple to return a
    deterministic regressor with the lowest empirical error subject to the
    constraint of bounded group loss [#agarwal19]_.

    References:
        .. [#agarwal18] `A. Agarwal, A. Beygelzimer, M. Dudik, J. Langford, and
           H. Wallach, "A Reductions Approach to Fair Classification,"
           International Conference on Machine Learning, 2018.
           <https://arxiv.org/abs/1803.02453>`_
        .. [#agarwal19] `A. Agarwal, M. Dudik, and Z. Wu, "Fair Regression:
           Quantitative Definitions and Reduction-based Algorithms,"
           International Conference on Machine Learning, 2019.
           <https://arxiv.org/abs/1905.12843>`_
    """
    def __init__(self,
                 estimator,
                 constraints,
                 prot_attr=None,
                 constraint_weight=0.5,
                 grid_size=10,
                 grid_limit=2.0,
                 grid=None,
                 drop_prot_attr=True,
                 loss="ZeroOne",
                 min_val=None,
                 max_val=None):
        """
        Args:
            estimator: An estimator implementing methods ``fit(X, y,
                sample_weight)`` and ``predict(X)``, where ``X`` is the matrix
                of features, ``y`` is the vector of labels, and
                ``sample_weight`` is a vector of weights; labels ``y`` and
                predictions returned by ``predict(X)`` are either 0 or 1 -- e.g.
                scikit-learn classifiers/regressors.
            constraints (str or fairlearn.reductions.Moment): If string, keyword
                denoting the :class:`fairlearn.reductions.Moment` object
                defining the disparity constraints -- e.g., "DemographicParity"
                or "EqualizedOdds". For a full list of possible options see
                `self.model.moments`. Otherwise, provide the desired
                :class:`~fairlearn.reductions.Moment` object defining the
                disparity constraints.
            prot_attr: String or array-like column indices or column names
                of protected attributes.
            constraint_weight: When the ``selection_rule`` is
                "tradeoff_optimization" (default, no other option currently)
                this float specifies the relative weight put on the constraint
                violation when selecting the best model. The weight placed on
                the error rate will be ``1-constraint_weight``.
            grid_size (int): The number of Lagrange multipliers to generate in
                the grid.
            grid_limit (float): The largest Lagrange multiplier to generate. The
                grid will contain values distributed between ``-grid_limit`` and
                ``grid_limit`` by default.
            grid (pandas.DataFrame): Instead of supplying a size and limit for
                the grid, users may specify the exact set of Lagrange
                multipliers they desire using this argument in a DataFrame.
            drop_prot_attr (bool): Flag indicating whether to drop protected
                attributes from training data.
            loss (str): String identifying loss function for constraints.
                Options include "ZeroOne", "Square", and "Absolute."
            min_val: Loss function parameter for "Square" and "Absolute,"
                typically the minimum of the range of y values.
            max_val: Loss function parameter for "Square" and "Absolute,"
                typically the maximum of the range of y values.
        """
        super(GridSearchReduction, self).__init__()

        #init model, set prot_attr during fit
        if prot_attr is None:
            prot_attr = []
        self.model = skGridSearchRed(prot_attr, estimator, constraints,
                constraint_weight, grid_size, grid_limit, grid, drop_prot_attr,
                loss, min_val, max_val)


    def fit(self, dataset):
        """Learns model with less bias

        Args:
            dataset : Dataset containing true output.

        Returns:
            GridSearchReduction: Returns self.
        """
        #set prot_attr
        if len(self.model.prot_attr) == 0:
            self.model.prot_attr = dataset.protected_attribute_names

        X_df = pd.DataFrame(dataset.features, columns=dataset.feature_names)
        Y = dataset.labels

        self.model.fit(X_df, Y)

        return self


    def predict(self, dataset):
        """Obtain the predictions for the provided dataset using the model
        learned.

        Args:
            dataset: Dataset containing output values that need to be
                transformed.

        Returns:
            dataset: Transformed dataset.
        """
        X_df = pd.DataFrame(dataset.features, columns=dataset.feature_names)

        dataset_new = dataset.copy()
        dataset_new.labels = self.model.predict(X_df).reshape(-1, 1)

        if isinstance(self.model.moment_, red.ClassificationMoment):
            fav = int(dataset.favorable_label)
            try:
                # Probability of favorable label
                scores = self.model.predict_proba(X_df)[:, fav]
                dataset_new.scores = scores.reshape(-1, 1)
            except (AttributeError, NotImplementedError):
                warning("dataset.scores not updated, underlying model does not "
                        "support predict_proba")

        return dataset_new



import numpy as np
from sklearn.base import BaseEstimator, MetaEstimatorMixin, clone
from sklearn.utils.metaestimators import available_if
from sklearn.utils.validation import has_fit_parameter

from aif360.sklearn.utils import check_inputs, check_groups


class Reweighing(BaseEstimator):
    """Sample reweighing.

    Reweighing is a preprocessing technique that weights the examples in each
    (group, label) combination differently to ensure fairness before
    classification [#kamiran12]_.

    Note:
        This breaks the scikit-learn API by returning new sample weights from
        ``fit_transform()``. See :class:`ReweighingMeta` for a workaround.

    See also:
        :class:`ReweighingMeta`

    References:
        .. [#kamiran12] `F. Kamiran and T. Calders,  "Data Preprocessing
           Techniques for Classification without Discrimination," Knowledge and
           Information Systems, 2012.
           <https://link.springer.com/article/10.1007/s10115-011-0463-8>`_

    Attributes:
        prot_attr_ (str or list(str)): Protected attribute(s) used for
            reweighing.
        groups_ (array, shape (n_groups,)): A list of group labels known to the
            transformer.
        classes_ (array, shape (n_classes,)): A list of class labels known to
            the transformer.
        reweigh_factors_ (array, shape (n_groups, n_labels)): Reweighing factors
            for each combination of group and class labels used to debias
            samples. Existing sample weights are multiplied by the corresponding
            factor for that sample's group and class.
    """

    def __init__(self, prot_attr=None):
        """
        Args:
            prot_attr (single label or list-like, optional): Protected
                attribute(s) to use in the reweighing process. If more than one
                attribute, all combinations of values (intersections) are
                considered. Default is ``None`` meaning all protected attributes
                from the dataset are used.
        """
        self.prot_attr = prot_attr

    def fit(self, X, y, sample_weight=None):
        """Only :meth:`fit_transform` is allowed for this algorithm."""
        self.fit_transform(X, y, sample_weight=sample_weight)
        return self

    def fit_transform(self, X, y, sample_weight=None):
        """Compute the factors for reweighing the dataset and transform the
        sample weights.

        Args:
            X (pandas.DataFrame): Training samples.
            y (array-like): Training labels.
            sample_weight (array-like, optional): Sample weights.

        Returns:
            tuple:
                Samples and their weights.

                * **X** -- Unchanged samples.
                * **sample_weight** -- Transformed sample weights.
        """
        X, y, sample_weight = check_inputs(X, y, sample_weight)

        sample_weight_t = np.empty_like(sample_weight, dtype=float)
        groups, self.prot_attr_ = check_groups(X, self.prot_attr)
        # TODO: maintain categorical ordering
        self.groups_ = np.unique(groups)
        self.classes_ = np.unique(y)
        n_groups = len(self.groups_)
        n_classes = len(self.classes_)
        self.reweigh_factors_ = np.full((n_groups, n_classes), np.nan)

        def N_(i): return sample_weight[i].sum()
        N = sample_weight.sum()
        for i, g in enumerate(self.groups_):
            for j, c in enumerate(self.classes_):
                g_and_c = (groups == g) & (y == c)
                if np.any(g_and_c):
                    W_gc = N_(groups == g) / N * N_(y == c) / N_(g_and_c)
                    sample_weight_t[g_and_c] = W_gc * sample_weight[g_and_c]
                    self.reweigh_factors_[i, j] = W_gc
        return X, sample_weight_t


class ReweighingMeta(BaseEstimator, MetaEstimatorMixin):
    """A meta-estimator which wraps a given estimator with a reweighing
    preprocessing step.

    This is necessary for use in a Pipeline, etc.

    Attributes:
        estimator_ (sklearn.BaseEstimator): The fitted underlying estimator.
        reweigher_: The fitted underlying reweigher.
        classes_ (array, shape (n_classes,)): Class labels from `estimator_`.
    """
    def __init__(self, estimator, reweigher=None):
        """
        Args:
            estimator (sklearn.BaseEstimator): Estimator to be wrapped.
            reweigher (optional): Preprocessor which returns new sample weights
                from ``transform()``. If ``None``, defaults to
                :class:`~aif360.sklearn.preprocessing.Reweighing`.
        """
        self.reweigher = reweigher
        self.estimator = estimator

    @property
    def _estimator_type(self):
        return self.estimator._estimator_type

    @property
    def classes_(self):
        """Class labels from the base estimator."""
        return self.estimator_.classes_

    def fit(self, X, y, sample_weight=None):
        """Performs ``self.reweigher_.fit_transform(X, y, sample_weight)`` and
        then ``self.estimator_.fit(X, y, sample_weight)`` using the reweighed
        samples.

        Args:
            X (pandas.DataFrame): Training samples.
            y (array-like): Training labels.
            sample_weight (array-like, optional): Sample weights.

        Returns:
            self
        """
        if not has_fit_parameter(self.estimator, 'sample_weight'):
            raise TypeError("`estimator` (type: {}) does not have fit parameter"
                            " `sample_weight`.".format(type(self.estimator)))

        if self.reweigher is None:
            self.reweigher_ = Reweighing()
        else:
            self.reweigher_ = clone(self.reweigher)
        self.estimator_ = clone(self.estimator)

        X, sample_weight = self.reweigher_.fit_transform(X, y,
                sample_weight=sample_weight)
        self.estimator_.fit(X, y, sample_weight=sample_weight)
        return self

    @available_if(lambda self: hasattr(self.estimator_, "predict"))
    def predict(self, X):
        """Predict class labels for the given samples using ``self.estimator_``.

        Args:
            X (array-like): Test samples.

        Returns:
            array: Predicted class label per sample.
        """
        return self.estimator_.predict(X)

    @available_if(lambda self: hasattr(self.estimator_, "predict_proba"))
    def predict_proba(self, X):
        """Probability estimates from ``self.estimator_``.

        The returned estimates for all classes are ordered by the label of
        classes.

        Args:
            X (array-like): Test samples.

        Returns:
            array: Returns the probability of the sample for each class in the
            model, where classes are ordered as they are in ``self.classes_``.
        """
        return self.estimator_.predict_proba(X)

    @available_if(lambda self: hasattr(self.estimator_, "predict_log_proba"))
    def predict_log_proba(self, X):
        """Log of probability estimates from ``self.estimator_``.

        The returned estimates for all classes are ordered by the label of
        classes.

        Args:
            X (array-like): Test samples.

        Returns:
            array: Returns the log-probability of the sample for each class in
            the model, where classes are ordered as they are in
            ``self.classes_``.
        """
        return self.estimator_.predict_log_proba(X)

    @available_if(lambda self: hasattr(self.estimator_, "score"))
    def score(self, X, y, sample_weight=None):
        """Returns the output of the estimator's score function on the given
        test data and labels.

        Args:
            X (array-like): Test samples.
            y (array-like): True labels for X.
            sample_weight (array-like, optional): Sample weights.

        Returns:
            float: `self.estimator.score(X, y, sample_weight)`
        """
        return self.estimator_.score(X, y, sample_weight=sample_weight)
